# -*- coding: utf-8 -*-
# Author: Dr. Martin Wolf <mail@martin-wolf.org>

"""This module contains utility functions for creating and managing
NDPhotosplinePDF instances.
"""

from skyllh.core.binning import BinningDefinition
from skyllh.core.pdf import NDPhotosplinePDF
from skyllh.core.backgroundpdf import BackgroundNDPhotosplinePDF
from skyllh.core.signalpdf import SignalNDPhotosplinePDF


def create_NDPhotosplinePDF_from_photosplinefit(
        ds, kind, info_key, splinefit_key, param_set=None,
        norm_factor_func=None,
        tl=None
    ):
    """Creates a new NDPhotosplinePDF instance from a photospline fits file that
    is defined in the given data set.

    Parameters
    ----------
    ds : Dataset instance
        The Dataset instance the PDF applies to.
    kind : str | None
        The kind of PDF to create. This is either ``'sig'`` for a
        SignalNDPhotosplinePDF, or ``'bkg'`` for a BackgroundNDPhotosplinePDF
        instance. If set to None, a NDPhotosplinePDF instance is created.
    info_key : str
        The auxiliary data name for the file containing PDF meta data
        information.
    splinefit_key : str
        The auxiliary data name defining the path to the file containing the
        photospline fit.
    param_set : Parameter | ParameterSet | None
        The Parameter instance or ParameterSet instance defining the
        parameters of the new PDF. The ParameterSet holds the information
        which parameters are fixed and which are floating (i.e. fitted).
    norm_factor_func : callable | None
        The normalization factor function. It must have the following call
        signature:
            __call__(pdf, tdm, params)
    tl : TimeLord instance | None
        The optional TimeLord instance to use for measuring timing information.

    Returns
    -------
    pdf : SignalNDPhotosplinePDF instance |
          BackgroundNDPhotosplinePDF instance | NDPhotosplinePDF instance
        The created PDF instance. Depending on the ``kind`` argument, this is
        a SignalNDPhotosplinePDF, a BackgroundNDPhotosplinePDF, or a
        NDPhotosplinePDF instance.
    """

    if(kind is None):
        pdf_type = NDPhotosplinePDF
    elif(kind == 'sig'):
        pdf_type = SignalNDPhotosplinePDF
    elif(kind == 'bkg'):
        pdf_type = BackgroundNDPhotosplinePDF
    else:
        raise ValueError(
            'The kind argument must be None, "sig", or "bkg"! '
            'Currently it is '+str(kind)+'!')

    # Load the PDF data from the auxilary files.
    info_dict = ds.load_aux_data(info_key, tl=tl)

    kde_pdf_axis_name_map = ds.load_aux_data('KDE_PDF_axis_name_map', tl=tl)
    kde_pdf_axis_name_map_inv = dict(
        [(v, k) for (k, v) in kde_pdf_axis_name_map.items()])
    for var in info_dict['vars']:
        if(var not in kde_pdf_axis_name_map_inv):
            kde_pdf_axis_name_map_inv[var] = var

    # Select the bin center information from the meta data information file.
    # The "bins" key is for backward compatibility.
    if('bin_centers' in info_dict):
        bin_centers_key = 'bin_centers'
    elif('bins' in info_dict):
        bin_centers_key = 'bins'
    else:
        raise KeyError(
            'The PDF information file is missing "bin_centers" or "bins" key!')

    axis_binnings = [
        BinningDefinition(
            kde_pdf_axis_name_map_inv[var], info_dict[bin_centers_key][idx])
        for (idx, var) in enumerate(info_dict['vars'])
    ]

    # Getting the name of the splinetable file
    splinefit_file_list = ds.get_abs_pathfilename_list(
        ds.get_aux_data_definition(splinefit_key))
    if(len(splinefit_file_list) != 1):
        raise ValueError(
            'The spline fit file list must contain only a single file name! '
            'Currently it contains {} file names!'.format(
                len(splinefit_file_list)))

    pdf = pdf_type(
        axis_binnings=axis_binnings,
        param_set=param_set,
        path_to_pdf_splinefit=splinefit_file_list[0],
        norm_factor_func=norm_factor_func)

    return pdf



# -*- coding: utf-8 -*-
# Author: Dr. Martin Wolf <mail@martin-wolf.org>

import unittest

from skyllh.core.py import (
    ConstPyQualifier,
    const,
    issequenceof
)


class A(object):
    def __init__(self, name=None):
        super(A, self).__init__()

        self.name = name

class B(object):
    def __init__(self, name=None):
        super(B, self).__init__()

        self.name = name


class ConstPyQualifier_TestCase(unittest.TestCase):
    def test_call(self):
        a = const(A())
        self.assertTrue(hasattr(a, '__pyqualifiers__'))
        self.assertTrue(ConstPyQualifier in a.__pyqualifiers__)

    def test_check(self):
        a = const(A())
        self.assertTrue(const.check(a))


class issequenceof_TestCase(unittest.TestCase):
    def test_type(self):
        seq = [A('a1'), A('a2')]
        self.assertTrue(issequenceof(seq, A))

        seq = [A('a1'), B('b1')]
        self.assertFalse(issequenceof(seq, A))

    def test_pyqualifiers(self):
        """Tests if the issequenceof function works with PyQualifiers.
        """
        seq = [const(A('a1')), const(A('a2'))]
        self.assertTrue(issequenceof(seq, A, const))

        seq = [const(A('a1')), A('a2')]
        self.assertFalse(issequenceof(seq, A, const))


if(__name__ == '__main__'):
    unittest.main()



# -*- coding: utf-8 -*-
# Author: Dr. Martin Wolf <mail@martin-wolf.org>

import numpy as np

def isAlmostEqual(a, b, decimals=9):
    a = np.atleast_1d(a)
    b = np.atleast_1d(b)
    return np.all(np.around(np.abs(a - b), decimals) == 0)



# -*- coding: utf-8 -*-

"""The analysis module provides classes for pre-defined analyses.
"""

import abc
import numpy as np
import pickle

from skyllh.core.py import (
    classname,
    issequenceof
)
from skyllh.core.debugging import get_logger
from skyllh.core.storage import DataFieldRecordArray
from skyllh.core.dataset import (
    Dataset,
    DatasetData
)
from skyllh.core.parameters import (
    FitParameter,
    SourceFitParameterMapper,
    SingleSourceFitParameterMapper
)
from skyllh.core.pdf import (
    EnergyPDF,
    SpatialPDF
)
from skyllh.core.pdfratio import PDFRatio
from skyllh.core.progressbar import ProgressBar
from skyllh.core.random import RandomStateService
from skyllh.core.llhratio import (
    LLHRatio,
    MultiDatasetTCLLHRatio,
    SingleSourceDatasetSignalWeights,
    SingleSourceZeroSigH0SingleDatasetTCLLHRatio,
    MultiSourceZeroSigH0SingleDatasetTCLLHRatio,
    MultiSourceDatasetSignalWeights
)
from skyllh.core.scrambling import DataScramblingMethod
from skyllh.core.timing import TaskTimer
from skyllh.core.trialdata import TrialDataManager
from skyllh.core.optimize import (
    EventSelectionMethod,
    AllEventSelectionMethod
)
from skyllh.core.source_hypothesis import SourceHypoGroupManager
from skyllh.core.test_statistic import TestStatistic
from skyllh.core.multiproc import get_ncpu, parallelize
from skyllh.core.background_generation import BackgroundGenerationMethod
from skyllh.core.background_generator import BackgroundGenerator
from skyllh.core.signal_generator import (
    SignalGeneratorBase,
    SignalGenerator
)
from skyllh.physics.source import SourceModel


logger = get_logger(__name__)


class Analysis(object, metaclass=abc.ABCMeta):
    """This is the abstract base class for all analysis classes. It contains
    common properties required by all analyses and defines the overall analysis
    interface howto set-up and run an analysis.

    Note: This analysis base class assumes the analysis to be a log-likelihood
          ratio test, i.e. requires a mathematical log-likelihood ratio
          function.

    To set-up and run an analysis the following procedure applies:

        1. Create an analysis instance.
        2. Add the datasets and their PDF ratio instances via the
           :meth:`.add_dataset` method.
        3. Construct the log-likelihood ratio function via the
           :meth:`.construct_llhratio` method.
        4. Call the :meth:`do_trial` or :meth:`unblind` method to perform a
           random trial or to unblind the data. Both methods will fit the global
           fit parameters using the set up data. Finally, the test statistic
           is calculated via the :meth:`.calculate_test_statistic` method.

    In order to calculate sensitivities and discovery potentials, analysis
    trials have to be performed on random data samples with injected signal
    events. To perform a trial with injected signal events, the signal generator
    has to be constructed via the ``construct_signal_generator`` method before
    any random trial data can be generated.
    """

    def __init__(self, src_hypo_group_manager, src_fitparam_mapper,
                 test_statistic, bkg_gen_method=None, sig_generator_cls=None):
        """Constructor of the analysis base class.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses, their flux model, and their detector signal
            efficiency implementation method.
        src_fitparam_mapper : instance of SourceFitParameterMapper
            The SourceFitParameterMapper instance managing the global fit
            parameters and their relation to the individual sources.
        test_statistic : TestStatistic instance
            The TestStatistic instance that defines the test statistic function
            of the analysis.
        bkg_gen_method : instance of BackgroundGenerationMethod | None
            The instance of BackgroundGenerationMethod that should be used to
            generate background events for pseudo data. This can be set to None,
            if there is no need to generate background events.
        sig_generator_cls : SignalGeneratorBase class | None
            The signal generator class used to create the signal generator
            instance.
            If set to None, the `SignalGenerator` class is used.
        """
        # Call the super function to allow for multiple class inheritance.
        super(Analysis, self).__init__()

        self.src_hypo_group_manager = src_hypo_group_manager
        self.src_fitparam_mapper = src_fitparam_mapper
        self.test_statistic = test_statistic
        self.bkg_gen_method = bkg_gen_method
        self.sig_generator_cls = sig_generator_cls

        self._dataset_list = []
        self._data_list = []
        self._tdm_list = []
        self._event_selection_method_list = []

        # Predefine the variable for the global fit parameter set, which holds
        # all the global fit parameters.
        self._fitparamset = None

        # Predefine the variable for the log-likelihood ratio function.
        self._llhratio = None

        # Predefine the variable for the background and signal generators.
        self._bkg_generator = None
        self._sig_generator = None

    @property
    def src_hypo_group_manager(self):
        """The SourceHypoGroupManager instance, which defines the groups of
        source hypothesis, their flux model, and their detector signal
        efficiency implementation method.
        """
        return self._src_hypo_group_manager
    @src_hypo_group_manager.setter
    def src_hypo_group_manager(self, manager):
        if(not isinstance(manager, SourceHypoGroupManager)):
            raise TypeError('The src_hypo_group_manager property must be an '
                'instance of SourceHypoGroupManager!')
        self._src_hypo_group_manager = manager

    @property
    def src_fitparam_mapper(self):
        """The SourceFitParameterMapper instance that manages the global fit
        parameters and their relation to the sources.
        """
        return self._src_fitparam_mapper
    @src_fitparam_mapper.setter
    def src_fitparam_mapper(self, mapper):
        if(not isinstance(mapper, SourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper property must be an '
                'instance of SourceFitParameterMapper!')
        self._src_fitparam_mapper = mapper

    @property
    def test_statistic(self):
        """The TestStatistic instance that defines the test-statistic function
        of the analysis.
        """
        return self._test_statistic
    @test_statistic.setter
    def test_statistic(self, ts):
        if(not isinstance(ts, TestStatistic)):
            raise TypeError('The test_statistic property must be an instance '
                'of TestStatistic, but is %s!'%(classname(ts)))
        self._test_statistic = ts

    @property
    def bkg_gen_method(self):
        """The BackgroundGenerationMethod instance that implements the
        background event generation. This can be None if no background
        generation method has been defined.
        """
        return self._bkg_gen_method
    @bkg_gen_method.setter
    def bkg_gen_method(self, method):
        if(method is not None):
            if(not isinstance(method, BackgroundGenerationMethod)):
                raise TypeError('The bkg_gen_method property must be an '
                    'instance of BackgroundGenerationMethod!')
        self._bkg_gen_method = method

    @property
    def dataset_list(self):
        """The list of Dataset instances.
        """
        return self._dataset_list
    @dataset_list.setter
    def dataset_list(self, datasets):
        if(not issequenceof(datasets, Dataset)):
            raise TypeError('The dataset_list property must be a sequence '
                'of Dataset instances!')
        self._dataset_list = list(datasets)

    @property
    def data_list(self):
        """The list of DatasetData instances holding the original data of the
        dataset.
        """
        return self._data_list
    @data_list.setter
    def data_list(self, datas):
        if(not issequenceof(datas, DatasetData)):
            raise TypeError('The data_list property must be a sequence '
                'of DatasetData instances!')
        self._data_list = list(datas)

    @property
    def n_datasets(self):
        """(read-only) The number of datasets used in this analysis.
        """
        return len(self._dataset_list)

    @property
    def fitparamset(self):
        """(read-only) The instance of FitParameterSet holding all the global
        fit parameters of the log-likelihood ratio function.
        """
        return self._fitparamset

    @property
    def llhratio(self):
        """The log-likelihood ratio function instance. It is None, if it has
        not been constructed yet.
        """
        if(self._llhratio is None):
            raise RuntimeError('The log-likelihood ratio function is not '
                'defined yet. Call the construct_analysis method first!')
        return self._llhratio
    @llhratio.setter
    def llhratio(self, obj):
        if(not isinstance(obj, LLHRatio)):
            raise TypeError('The llhratio property must be an instance of '
                'LLHRatio!')
        self._llhratio = obj

    @property
    def bkg_generator(self):
        """(read-only) The background generator instance. Is None of the
        background generator has not been constructed via the
        `construct_background_generator` method.
        """
        return self._bkg_generator

    @property
    def sig_generator_cls(self):
        """The signal generator class that should be used to construct the
        signal generator instance.
        """
        return self._sig_generator_cls
    @sig_generator_cls.setter
    def sig_generator_cls(self, cls):
        if cls is None:
            cls = SignalGenerator
        if not issubclass(cls, SignalGeneratorBase):
            raise TypeError(
                'The sig_generator_cls property must be an subclass of '
                'SignalGeneratorBase!')
        self._sig_generator_cls = cls

    @property
    def sig_generator(self):
        """(read-only) The signal generator instance. Is None if the signal
        generator has not been constructed via the
        `construct_signal_generator` method.
        """
        return self._sig_generator

    @property
    def total_livetime(self):
        """(read-only) The total live-time in days of the loaded data.
        """
        livetime = 0
        for data in self._data_list:
            livetime += data.livetime
        return livetime

    def add_dataset(self, dataset, data, tdm=None, event_selection_method=None):
        """Adds the given dataset to the list of datasets for this analysis.

        Parameters
        ----------
        dataset : Dataset instance
            The Dataset instance that should get added.
        data : DatasetData instance
            The DatasetData instance holding the original (prepared) data of the
            dataset.
        tdm : TrialDataManager instance | None
            The TrialDataManager instance managing the trial data and additional
            data fields of the data set. If set to None, it means that no
            additional data fields are defined.
        event_selection_method : instance of EventSelectionMethod | None
            The instance of EventSelectionMethod to use to select only
            signal-like events from the data. All other events
            will be treated as pure background events. This reduces the amount
            of log-likelihood-ratio function evaluations. If set to None, all
            events will be evaluated.
        """
        if(not isinstance(dataset, Dataset)):
            raise TypeError(
                'The dataset argument must be an instance of Dataset!')

        if(not isinstance(data, DatasetData)):
            raise TypeError(
                'The data argument must be an instance of DatasetData!')

        if(tdm is None):
            tdm = TrialDataManager()
        if(not isinstance(tdm, TrialDataManager)):
            raise TypeError(
                'The tdm argument must be None or an instance of '
                'TrialDataManager!')

        if(event_selection_method is not None):
            if(not isinstance(event_selection_method, EventSelectionMethod)):
                raise TypeError(
                    'The event_selection_method argument must be None or an '
                    'instance of EventSelectionMethod!')

        self._dataset_list.append(dataset)
        self._data_list.append(data)
        self._tdm_list.append(tdm)
        self._event_selection_method_list.append(event_selection_method)

    def calculate_test_statistic(
            self, log_lambda, fitparam_values, *args, **kwargs):
        """Calculates the test statistic value by calling the ``evaluate``
        method of the TestStatistic class with the given log_lambda value and
        fit parameter values.

        Parameters
        ----------
        log_lambda : float
            The value of the log-likelihood ratio function. Usually, this is its
            maximum.
        fitparam_values : (N_fitparam+1)-shaped 1D ndarray
            The 1D ndarray holding the fit parameter values of the
            log-likelihood ratio function for the given log_lambda value.

        Additional arguments and keyword arguments
        ------------------------------------------
        Any additional arguments and keyword arguments are passed to the
        evaluate method of the TestStatistic class instance.

        Returns
        -------
        TS : float
            The calculated test-statistic value.
        """
        return self._test_statistic.evaluate(
            self._llhratio, log_lambda, fitparam_values, *args, **kwargs)

    @abc.abstractmethod
    def construct_llhratio(self):
        """This method is supposed to construct the log-likelihood ratio
        function and sets it as the _llhratio property.
        """
        pass

    def construct_background_generator(self):
        """Constructs the background generator for all added datasets.
        This method must be called after all the datasets were added via the
        add_dataset method. It sets the `bkg_generator` property of this
        Analysis class instance.
        """
        if(self._bkg_gen_method is None):
            raise RuntimeError('No background generation method has been '
                'defined for this analysis!')

        self._bkg_generator = BackgroundGenerator(
            self._bkg_gen_method, self._dataset_list, self._data_list)

    def construct_signal_generator(self):
        """Constructs the signal generator for all added datasets.
        This method must be called after all the datasets were added via the
        add_dataset method. It sets the `sig_generator` property of this
        Analysis class instance. The signal generation method has to be set
        through the source hypothesis group.
        """
        self._sig_generator = self.sig_generator_cls(
            src_hypo_group_manager=self._src_hypo_group_manager,
            dataset_list=self._dataset_list,
            data_list=self._data_list)

    @abc.abstractmethod
    def initialize_trial(self, events_list, n_events_list=None):
        """This method is supposed to initialize the log-likelihood ratio
        function with a new set of given trial data. This is a low-level method.
        For convenient methods see the `unblind` and `do_trial` methods.

        Parameters
        ----------
        events_list : list of numpy record ndarray
            The list of data events to use for the log-likelihood function
            evaluation. The data arrays for the datasets must be in the same
            order than the added datasets.
        n_events_list : list of int | None
            The list of the number of events of each data set. These numbers
            can be larger than the number of events given by the `events_list`
            argument in cases where an event selection method was already used.
            If set to None, the number of events is taken from the given
            `events_list` argument.
        """
        pass

    @abc.abstractmethod
    def maximize_llhratio(self, rss, tl=None):
        """This method is supposed to maximize the log-likelihood ratio
        function, by calling the ``maximize`` method of the LLHRatio class.

        Parameters
        ----------
        rss : RandomStateService instance
            The RandomStateService instance to draw random numbers from.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to time the
            maximization of the LLH ratio function.

        Returns
        -------
        fitparamset : FitParameterSet instance
            The instance of FitParameterSet holding the global fit parameter
            definitions used in the maximization process.
        log_lambda_max : float
            The value of the log-likelihood ratio function at its maximum.
        fitparam_values : (N_fitparam,)-shaped 1D ndarray
            The ndarray holding the global fit parameter values.
            By definition, the first element is the value of the fit parameter
            ns.
        status : dict
            The dictionary with status information about the maximization
            process, i.e. from the minimizer.
        """
        pass

    def unblind(self, rss):
        """Evaluates the unscrambled data, i.e. unblinds the data.

        Parameters
        ----------
        rss : RandomStateService instance
            The RandomStateService instance that should be used draw random
            numbers from.

        Returns
        -------
        TS : float
            The test-statistic value.
        fitparam_dict : dict
            The dictionary holding the global fit parameter names and their best
            fit values.
        status : dict
            The status dictionary with information about the performed
            minimization process of the negative of the log-likelihood ratio
            function.
        """
        events_list = [ data.exp for data in self._data_list ]
        self.initialize_trial(events_list)
        (fitparamset, log_lambda_max, fitparam_values, status) = self.maximize_llhratio(rss)
        TS = self.calculate_test_statistic(log_lambda_max, fitparam_values)

        fitparam_dict = fitparamset.fitparam_values_to_dict(fitparam_values)

        return (TS, fitparam_dict, status)

    def generate_background_events(
            self, rss, mean_n_bkg_list=None, bkg_kwargs=None, tl=None):
        """Generates background events utilizing the background generator.

        Parameters
        ----------
        rss : RandomStateService
            The RandomStateService instance to use for generating random
            numbers.
        mean_n_bkg_list : list of float | None
            The mean number of background events that should be generated for
            each dataset. If set to None (the default), the background
            generation method needs to obtain this number itself.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks of this method.

        Returns
        -------
        n_events_list : list of int
            The list of the number of events that have been generated for each
            pseudo data set.
        events_list : list of DataFieldRecordArray instances
            The list of DataFieldRecordArray instances containing the pseudo
            data events for each data sample. The number of events for each
            data set can be less than the number of events given by
            `n_events_list` if an event selection method was already utilized
            when generating background events.
        """
        n_datasets = self.n_datasets

        if(not isinstance(rss, RandomStateService)):
            raise TypeError(
                'The rss argument must be an instance of RandomStateService!')

        if(mean_n_bkg_list is None):
            mean_n_bkg_list = [ None ] * n_datasets
        if(not issequenceof(mean_n_bkg_list, (type(None), float))):
            raise TypeError(
                'The mean_n_bkg_list argument must be a sequence of None '
                'and/or floats!')

        if(bkg_kwargs is None):
            bkg_kwargs = dict()

        # Construct the background event generator in case it's not constructed
        # yet.
        if(self._bkg_generator is None):
            self.construct_background_generator()

        n_events_list = []
        events_list = []
        for ds_idx in range(n_datasets):
            bkg_kwargs.update(mean=mean_n_bkg_list[ds_idx])
            with TaskTimer(tl, 'Generating background events for data set '
                '{:d}.'.format(ds_idx)):
                (n_bkg, bkg_events) = self._bkg_generator.generate_background_events(
                    rss, ds_idx, tl=tl, **bkg_kwargs)
            n_events_list.append(n_bkg)
            events_list.append(bkg_events)

        return (n_events_list, events_list)

    def generate_signal_events(
            self, rss, mean_n_sig, sig_kwargs=None, n_events_list=None,
            events_list=None, tl=None):
        """Generates signal events utilizing the signal generator.

        Parameters
        ----------
        rss : RandomStateService
            The RandomStateService instance to use for generating random
            numbers.
        mean_n_sig : float
            The mean number of signal events that should be generated for the
            trial. The actual number of generated events will be drawn from a
            Poisson distribution with this given signal mean as mean.
        sig_kwargs : dict | None
            Additional keyword arguments for the `generate_signal_events` method
            of the `sig_generator_cls` class. An usual keyword argument is
            `poisson`.
        n_events_list : list of int | None
            If given, it specifies the number of events of each data set already
            present and the number of signal events will be added.
        events_list : list of DataFieldRecordArray instances | None
            If given, it specifies the events of each data set already present
            and the signal events will be added.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks of this method.

        Returns
        -------
        n_sig : int
            The actual number of injected signal events.
        n_events_list : list of int
            The list of the number of events that have been generated for each
            pseudo data set.
        events_list : list of DataFieldRecordArray instances
            The list of DataFieldRecordArray instances containing the pseudo
            signal data events for each data set. An entry is None, if no signal
            events were generated for this particular data set.
        """
        n_datasets = self.n_datasets

        if(not isinstance(rss, RandomStateService)):
            raise TypeError(
                'The rss argument must be an instance of RandomStateService!')

        if(sig_kwargs is None):
            sig_kwargs = dict()

        if(n_events_list is None):
            n_events_list = [0] * n_datasets
        else:
            if(not issequenceof(n_events_list, int)):
                raise TypeError(
                    'The n_events_list argument must be a sequence of '
                    'instances of type int!')
            if(len(n_events_list) != n_datasets):
                raise ValueError(
                    'The n_events_list argument must be a list of int of '
                    'length {:d}! Currently it is of length {:d}.'.format(
                        n_datasets, len(n_events_list)))

        if(events_list is None):
            events_list = [None] * n_datasets
        else:
            if(not issequenceof(
                    events_list, (type(None), DataFieldRecordArray))):
                raise TypeError(
                    'The events_list argument must be a sequence of '
                    'instances of type DataFieldRecordArray!')
            if(len(events_list) != n_datasets):
                raise ValueError(
                    'The events_list argument must be a list of instances of '
                    'type DataFieldRecordArray with a length of {:d}! '
                    'Currently it is of length {:d}.'.format(
                        n_datasets, len(events_list)))

        n_sig = 0

        if(mean_n_sig == 0):
            return (n_sig, n_events_list, events_list)

        # Construct the signal generator if not done yet.
        if(self._sig_generator is None):
            with TaskTimer(tl, 'Constructing signal generator.'):
                self.construct_signal_generator()
        # Generate signal events with the given mean number of signal
        # events.
        sig_kwargs.update(mean=mean_n_sig)
        with TaskTimer(tl, 'Generating signal events.'):
            (n_sig, ds_sig_events_dict) = self._sig_generator.generate_signal_events(
                rss, **sig_kwargs)
        # Inject the signal events to the generated background data.
        for (ds_idx, sig_events) in ds_sig_events_dict.items():
            n_events_list[ds_idx] += len(sig_events)
            if(events_list[ds_idx] is None):
                events_list[ds_idx] = sig_events
            else:
                events_list[ds_idx].append(sig_events)

        return (n_sig, n_events_list, events_list)

    def generate_pseudo_data(
            self, rss, mean_n_bkg_list=None, mean_n_sig=0, bkg_kwargs=None,
            sig_kwargs=None, tl=None):
        """Generates pseudo data with background and possible signal
        events for each data set using the background and signal generation
        methods of the analysis.

        Parameters
        ----------
        rss : RandomStateService
            The RandomStateService instance to use for generating random
            numbers.
        mean_n_bkg_list : list of float | None
            The mean number of background events that should be generated for
            each dataset. If set to None (the default), the background
            generation method needs to obtain this number itself.
        mean_n_sig : float
            The mean number of signal events that should be generated for the
            trial. The actual number of generated events will be drawn from a
            Poisson distribution with this given signal mean as mean.
        bkg_kwargs : dict | None
            Additional keyword arguments for the `generate_events` method of the
            background generation method class. An usual keyword argument is
            `poisson`.
        sig_kwargs : dict | None
            Additional keyword arguments for the `generate_signal_events` method
            of the `SignalGenerator` class. An usual keyword argument is
            `poisson`.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks of this method.

        Returns
        -------
        n_sig : int
            The actual number of injected signal events.
        n_events_list : list of int
            The list of the number of events that have been generated for each
            pseudo data set.
        events_list : list of DataFieldRecordArray instances
            The list of DataFieldRecordArray instances containing the pseudo
            data events for each data sample. The number of events for each
            data set can be less than the number of events given by
            `n_events_list` if an event selection method was already utilized
            when generating background events.
        """
        # Generate background events for each dataset.
        (n_events_list, events_list) = self.generate_background_events(
            rss=rss,
            mean_n_bkg_list=mean_n_bkg_list,
            bkg_kwargs=bkg_kwargs,
            tl=tl)

        # Generate signal events and add them to the already generated
        # background events.
        (n_sig, n_events_list, events_list) = self.generate_signal_events(
            rss=rss,
            mean_n_sig=mean_n_sig,
            sig_kwargs=sig_kwargs,
            n_events_list=n_events_list,
            events_list=events_list,
            tl=tl)

        return (n_sig, n_events_list, events_list)

    def do_trial_with_given_pseudo_data(
            self, rss, mean_n_sig, n_sig, n_events_list, events_list,
            mean_n_sig_0=None,
            minimizer_status_dict=None,
            tl=None):
        """Performs an analysis trial on the given pseudo data.

        Parameters
        ----------
        rss : RandomStateService
            The RandomStateService instance to use for generating random
            numbers.
        mean_n_sig : float
            The mean number of signal events the pseudo data was generated with.
        n_sig : int
            The total number of actual signal events in the pseudo data.
        n_events_list : list of int
            The total number of events for each data set of the pseudo data.
        events_list : list of DataFieldRecordArray instances
            The list of DataFieldRecordArray instances containing the pseudo
            data events for each data sample. The number of events for each
            data sample can be less than the number of events given by
            `n_events_list` if an event selection method was already utilized
            when generating background events.
        mean_n_sig_0 : float | None
            The fixed mean number of signal events for the null-hypothesis,
            when using a ns-profile log-likelihood-ratio function.
            If set to None, this argument is interpreted as 0.
        minimizer_status_dict : dict | None
            If a dictionary is provided, it will be updated with the minimizer
            status dictionary.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks.

        Returns
        -------
        result : structured numpy ndarray
            The structured numpy ndarray holding the result of the trial. It
            contains the following data fields:

            rss_seed : int
                The RandomStateService seed.
            mean_n_sig : float
                The mean number of signal events.
            n_sig : int
                The actual number of injected signal events.
            mean_n_sig_0 : float
                The fixed mean number of signal events for the null-hypothesis.
            ts : float
                The test-statistic value.
            [<fitparam_name> ... : float ]
                Any additional fit parameters of the LLH function.
        """
        if(mean_n_sig_0 is not None):
            self._llhratio.mean_n_sig_0 = mean_n_sig_0
        else:
            mean_n_sig_0 = 0

        with TaskTimer(tl, 'Initializing trial.'):
            self.initialize_trial(events_list, n_events_list)

        with TaskTimer(tl, 'Maximizing LLH ratio function.'):
            (fitparamset, log_lambda_max, fitparam_values, status) = self.maximize_llhratio(rss, tl=tl)
        if(isinstance(minimizer_status_dict, dict)):
            minimizer_status_dict.update(status)

        with TaskTimer(tl, 'Calculating test statistic.'):
            ts = self.calculate_test_statistic(log_lambda_max, fitparam_values)

        # Create the structured array data type for the result array.
        result_dtype = [
            ('seed', np.int64),
            ('mean_n_sig', np.float64),
            ('n_sig', np.int64),
            ('mean_n_sig_0', np.float64),
            ('ts', np.float64)
        ] + [
            (fitparam_name, np.float64)
                for fitparam_name in fitparamset.fitparam_name_list
        ]
        result = np.empty((1,), dtype=result_dtype)
        result['seed'] = rss.seed
        result['mean_n_sig'] = mean_n_sig
        result['n_sig'] = n_sig
        result['mean_n_sig_0'] = mean_n_sig_0
        result['ts'] = ts
        for (idx, fitparam_name) in enumerate(fitparamset.fitparam_name_list):
            result[fitparam_name] = fitparam_values[idx]

        return result

    def do_trial_with_given_bkg_and_sig_pseudo_data(
            self, rss, mean_n_sig, n_sig, n_bkg_events_list, n_sig_events_list,
            bkg_events_list, sig_events_list,
            mean_n_sig_0=None,
            minimizer_status_dict=None,
            tl=None):
        """Performs an analysis trial on the given background and signal pseudo
        data. This method merges the background and signal pseudo events and
        calls the ``do_trial_with_given_pseudo_data`` method of this class.

        Note
        ----
        This method alters the DataFieldRecordArray instances of the
        bkg_events_list argument!

        Parameters
        ----------
        rss : RandomStateService
            The RandomStateService instance to use for generating random
            numbers.
        mean_n_sig : float
            The mean number of signal events the pseudo data was generated with.
        n_sig : int
            The total number of actual signal events in the pseudo data.
        n_bkg_events_list : list of int
            The total number of background events for each data set of the
            pseudo data.
        n_sig_events_list : list of int
            The total number of signal events for each data set of the
            pseudo data.
        bkg_events_list : list of DataFieldRecordArray instances
            The list of DataFieldRecordArray instances containing the background
            pseudo data events for each data set.
        sig_events_list : list of DataFieldRecordArray instances or None
            The list of DataFieldRecordArray instances containing the signal
            pseudo data events for each data set. If a particular dataset has
            no signal events, the entry for that dataset can be None.
        mean_n_sig_0 : float | None
            The fixed mean number of signal events for the null-hypothesis,
            when using a ns-profile log-likelihood-ratio function.
            If set to None, this argument is interpreted as 0.
        minimizer_status_dict : dict | None
            If a dictionary is provided, it will be updated with the minimizer
            status dictionary.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks.

        Returns
        -------
        result : structured numpy ndarray
            The structured numpy ndarray holding the result of the trial.
            See the documentation of the ``do_trial_with_given_pseudo_data``
            method for further information.
        """
        n_events_list = list(
            np.array(n_bkg_events_list) +
            np.array(n_sig_events_list)
        )

        events_list = bkg_events_list

        # Add potential signal events to the background events.
        for ds_idx in range(len(events_list)):
            if(sig_events_list[ds_idx] is not None):
                if(events_list[ds_idx] is None):
                    events_list[ds_idx] = sig_events_list[ds_idx]
                else:
                    events_list[ds_idx].append(sig_events_list[ds_idx])

        return self.do_trial_with_given_pseudo_data(
            rss = rss,
            mean_n_sig = mean_n_sig,
            n_sig = n_sig,
            n_events_list = n_events_list,
            events_list = events_list,
            mean_n_sig_0 = mean_n_sig_0,
            minimizer_status_dict = minimizer_status_dict,
            tl = tl
        )

    def do_trial(
            self, rss, mean_n_bkg_list=None, mean_n_sig=0, mean_n_sig_0=None,
            bkg_kwargs=None, sig_kwargs=None, minimizer_status_dict=None,
            tl=None):
        """Performs an analysis trial by generating a pseudo data sample with
        background events and possible signal events, and performs the LLH
        analysis on that random pseudo data sample.

        Parameters
        ----------
        rss : RandomStateService
            The RandomStateService instance to use for generating random
            numbers.
        mean_n_bkg_list : list of float | None
            The mean number of background events that should be generated for
            each dataset. If set to None (the default), the background
            generation method needs to obtain this number itself.
        mean_n_sig : float
            The mean number of signal events that should be generated for the
            trial. The actual number of generated events will be drawn from a
            Poisson distribution with this given signal mean as mean.
        mean_n_sig_0 : float | None
            The fixed mean number of signal events for the null-hypothesis,
            when using a ns-profile log-likelihood-ratio function.
            If set to None, this argument is interpreted as 0.
        bkg_kwargs : dict | None
            Additional keyword arguments for the `generate_events` method of the
            background generation method class. An usual keyword argument is
            `poisson`.
        sig_kwargs : dict | None
            Additional keyword arguments for the `generate_signal_events` method
            of the `SignalGenerator` class. An usual keyword argument is
            `poisson`.
        minimizer_status_dict : dict | None
            If a dictionary is provided, it will be updated with the minimizer
            status dictionary.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks.

        Returns
        -------
        result : structured numpy ndarray
            The structured numpy ndarray holding the result of the trial. It
            contains the following data fields:

            mean_n_sig : float
                The mean number of signal events.
            n_sig : int
                The actual number of injected signal events.
            mean_n_sig_0 : float
                The fixed mean number of signal events for the null-hypothesis.
            ts : float
                The test-statistic value.
            [<fitparam_name> ... : float ]
                Any additional fit parameters of the LLH function.
        """
        if(mean_n_sig_0 is not None):
            self._llhratio.mean_n_sig_0 = mean_n_sig_0
        else:
            mean_n_sig_0 = 0

        with TaskTimer(tl, 'Generating pseudo data.'):
            (n_sig, n_events_list, events_list) = self.generate_pseudo_data(
                rss=rss, mean_n_bkg_list=mean_n_bkg_list, mean_n_sig=mean_n_sig,
                bkg_kwargs=bkg_kwargs, sig_kwargs=sig_kwargs, tl=tl)

        result = self.do_trial_with_given_pseudo_data(
            rss=rss,
            mean_n_sig=mean_n_sig,
            n_sig=n_sig,
            n_events_list=n_events_list,
            events_list=events_list,
            mean_n_sig_0=mean_n_sig_0,
            minimizer_status_dict=minimizer_status_dict,
            tl=tl
        )

        return result

    def do_trials(
            self, rss, n, mean_n_bkg_list=None, mean_n_sig=0, mean_n_sig_0=None,
            bkg_kwargs=None, sig_kwargs=None, ncpu=None, tl=None, ppbar=None):
        """Executes `do_trial` method `N` times with possible multi-processing.
        One trial performs an analysis trial by generating a pseudo data sample
        with background events and possible signal events, and performs the LLH
        analysis on that random pseudo data sample.

        Parameters
        ----------
        rss : RandomStateService
            The RandomStateService instance to use for generating random
            numbers.
        n : int
            Number of trials to generate using the `do_trial` method.
        mean_n_bkg_list : list of float | None
            The mean number of background events that should be generated for
            each dataset. If set to None (the default), the number of data
            events of each data sample will be used as mean.
        mean_n_sig : float
            The mean number of signal events that should be generated for the
            trial. The actual number of generated events will be drawn from a
            Poisson distribution with this given signal mean as mean.
        mean_n_sig_0 : float | None
            The fixed mean number of signal events for the null-hypothesis,
            when using a ns-profile log-likelihood-ratio function.
        bkg_kwargs : dict | None
            Additional keyword arguments for the `generate_events` method of the
            background generation method class. An usual keyword argument is
            `poisson`.
        sig_kwargs : dict | None
            Additional keyword arguments for the `generate_signal_events` method
            of the `SignalGenerator` class. An usual keyword argument is
            `poisson`. If `poisson` is set to True, the actual number of
            generated signal events will be drawn from a Poisson distribution
            with the given mean number of signal events.
            If set to False, the argument ``mean_n_sig`` specifies the actual
            number of generated signal events.
        ncpu : int | None
            The number of CPUs to use, i.e. the number of subprocesses to
            spawn. If set to None, the global setting will be used.
        tl : instance of TimeLord | None
            The instance of TimeLord that should be used to time individual
            tasks.
        ppbar : instance of ProgressBar | None
            The possible parent ProgressBar instance.

        Returns
        -------
        result : structured numpy ndarray
            The structured numpy ndarray holding the result of the trial. It
            contains the following data fields:

            n_sig : int
                The actual number of injected signal events.
            ts : float
                The test-statistic value.
            [<fitparam_name> ... : float ]
                Any additional fit parameters of the LLH function.
        """
        ncpu = get_ncpu(ncpu)
        args_list = [((), {
            'mean_n_bkg_list': mean_n_bkg_list,
            'mean_n_sig': mean_n_sig,
            'mean_n_sig_0': mean_n_sig_0,
            'bkg_kwargs': bkg_kwargs,
            'sig_kwargs': sig_kwargs
            }) for i in range(n)
        ]
        result_list = parallelize(
            self.do_trial, args_list, ncpu, rss=rss, tl=tl, ppbar=ppbar)

        result_dtype = result_list[0].dtype
        result = np.empty(n, dtype=result_dtype)
        result[:] = result_list[:]
        return result


class TimeIntegratedMultiDatasetSingleSourceAnalysis(Analysis):
    """This is an analysis class that implements a time-integrated LLH ratio
    analysis for multiple datasets assuming a single source.

    To run this analysis the following procedure applies:

        1. Add the datasets and their spatial and energy PDF ratio instances
           via the :meth:`.add_dataset` method.
        2. Construct the log-likelihood ratio function via the
           :meth:`construct_llhratio` method.
        3. Initialize a trial via the :meth:`initialize_trial` method.
        4. Fit the global fit parameters to the trial data via the
           :meth:`maximize_llhratio` method.
    """
    def __init__(self, src_hypo_group_manager, src_fitparam_mapper, fitparam_ns,
                 test_statistic, bkg_gen_method=None, sig_generator_cls=None):
        """Creates a new time-integrated point-like source analysis assuming a
        single source.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses, their flux model, and their detector signal
            efficiency implementation method.
        src_fitparam_mapper : instance of SingleSourceFitParameterMapper
            The instance of SingleSourceFitParameterMapper defining the global
            fit parameters and their mapping to the source fit parameters.
        fitparam_ns : FitParameter instance
            The FitParameter instance defining the fit parameter ns.
        test_statistic : TestStatistic instance
            The TestStatistic instance that defines the test statistic function
            of the analysis.
        bkg_gen_method : instance of BackgroundGenerationMethod | None
            The instance of BackgroundGenerationMethod that will be used to
            generate background events for a new analysis trial. This can be set
            to None, if no background events have to get generated.
        sig_generator_cls : SignalGeneratorBase class | None
            The signal generator class used to create the signal generator
            instance.
            If set to None, the `SignalGenerator` class is used.
        """
        if(not isinstance(src_fitparam_mapper, SingleSourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper argument must be an '
                'instance of SingleSourceFitParameterMapper!')

        super().__init__(
            src_hypo_group_manager=src_hypo_group_manager,
            src_fitparam_mapper=src_fitparam_mapper,
            test_statistic=test_statistic,
            bkg_gen_method=bkg_gen_method,
            sig_generator_cls=sig_generator_cls)

        self.fitparam_ns = fitparam_ns

        # Define the member for the list of PDF ratio lists. Each list entry is
        # a list of PDF ratio instances for each data set.
        self._pdfratio_list_list = []

        # Create the FitParameterSet instance holding the fit parameter ns and
        # all the other additional fit parameters. This set is used by the
        # ``maximize_llhratio`` method.
        self._fitparamset = self._src_fitparam_mapper.fitparamset.copy()
        self._fitparamset.add_fitparam(self._fitparam_ns, atfront=True)

    @property
    def fitparam_ns(self):
        """The FitParameter instance for the fit parameter ns.
        """
        return self._fitparam_ns
    @fitparam_ns.setter
    def fitparam_ns(self, fitparam):
        if(not isinstance(fitparam, FitParameter)):
            raise TypeError('The fitparam_ns property must be an instance of FitParameter!')
        self._fitparam_ns = fitparam

    def add_dataset(self, dataset, data, pdfratios, tdm=None,
                    event_selection_method=None):
        """Adds a dataset with its PDF ratio instances to the analysis.

        Parameters
        ----------
        dataset : Dataset instance
            The Dataset instance that should get added.
        data : DatasetData instance
            The DatasetData instance holding the original (prepared) data of the
            dataset.
        pdfratios : PDFRatio instance | sequence of PDFRatio instances
            The PDFRatio instance or the sequence of PDFRatio instances for the
            to-be-added data set.
        tdm : TrialDataManager instance | None
            The TrialDataManager instance that manages the trial data and
            additional data fields for this data set.
        event_selection_method : instance of EventSelectionMethod | None
            The instance of EventSelectionMethod to use to select only
            signal-like events from the trial data. All other events
            will be treated as pure background events. This reduces the amount
            of log-likelihood-ratio function evaluations. If set to None, all
            events will be evaluated.
        """
        super(TimeIntegratedMultiDatasetSingleSourceAnalysis, self).add_dataset(
            dataset, data, tdm, event_selection_method)

        if(isinstance(pdfratios, PDFRatio)):
            pdfratios = [pdfratios]
        if(not issequenceof(pdfratios, PDFRatio)):
            raise TypeError('The pdfratios argument must be an instance of '
                'PDFRatio or a sequence of PDFRatio!')

        self._pdfratio_list_list.append(list(pdfratios))

    def construct_llhratio(self, minimizer, ppbar=None):
        """Constructs the log-likelihood-ratio (LLH-ratio) function of the
        analysis. This setups all the necessary analysis
        objects like detector signal efficiencies and dataset signal weights,
        constructs the log-likelihood ratio functions for each dataset and the
        final composite llh ratio function.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The instance of Minimizer that should be used to minimize the
            negative of the log-likelihood ratio function.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        llhratio : instance of MultiDatasetTCLLHRatio
            The instance of MultiDatasetTCLLHRatio that implements the
            log-likelihood-ratio function of the analysis.
        """
        # Create the detector signal yield instances for each dataset.
        # Since this is for a single source, we don't have to have individual
        # detector signal yield instances for each source as well.
        detsigyield_list = []
        fluxmodel = self._src_hypo_group_manager.get_fluxmodel_by_src_idx(0)
        detsigyield_implmethod_list = self._src_hypo_group_manager.get_detsigyield_implmethod_list_by_src_idx(0)
        if((len(detsigyield_implmethod_list) != 1) and
           (len(detsigyield_implmethod_list) != self.n_datasets)):
            raise ValueError('The number of detector signal yield '
                'implementation methods is not 1 and does not match the number '
                'of used datasets in the analysis!')
        pbar = ProgressBar(len(self.dataset_list), parent=ppbar).start()
        for (j, (dataset, data)) in enumerate(zip(self.dataset_list,
                                                  self.data_list)):
            if(len(detsigyield_implmethod_list) == 1):
                # Only one detsigyield implementation method was defined, so we
                # use it for all datasets.
                detsigyield_implmethod = detsigyield_implmethod_list[0]
            else:
                detsigyield_implmethod = detsigyield_implmethod_list[j]

            detsigyield = detsigyield_implmethod.construct_detsigyield(
                dataset, data, fluxmodel, data.livetime, ppbar=pbar)
            detsigyield_list.append(detsigyield)
            pbar.increment()
        pbar.finish()

        # For multiple datasets we need a dataset signal weights instance in
        # order to distribute ns over the different datasets.
        dataset_signal_weights = SingleSourceDatasetSignalWeights(
            self._src_hypo_group_manager, self._src_fitparam_mapper,
            detsigyield_list)

        # Create the list of log-likelihood ratio functions, one for each
        # dataset.
        llhratio_list = []
        for j in range(self.n_datasets):
            tdm = self._tdm_list[j]
            pdfratio_list = self._pdfratio_list_list[j]
            llhratio = SingleSourceZeroSigH0SingleDatasetTCLLHRatio(
                minimizer,
                self._src_hypo_group_manager,
                self._src_fitparam_mapper,
                tdm,
                pdfratio_list
            )
            llhratio_list.append(llhratio)

        # Create the final multi-dataset log-likelihood ratio function.
        llhratio = MultiDatasetTCLLHRatio(
            minimizer, dataset_signal_weights, llhratio_list)

        return llhratio

    def change_source(self, source):
        """Changes the source of the analysis to the given source. It makes the
        necessary changes to all the objects of the analysis.

        Parameters
        ----------
        source : SourceModel instance
            The SourceModel instance describing the new source.
        """
        if(not isinstance(source, SourceModel)):
            raise TypeError('The source argument must be an instance of SourceModel')

        if(self._llhratio is None):
            raise RuntimeError('The LLH ratio function has to be constructed, '
                'before the `change_source` method can be called!')

        # Change the source in the SourceHypoGroupManager instance.
        # Because this is a single source analysis, there can only be one source
        # hypothesis group defined.
        self._src_hypo_group_manager.src_hypo_group_list[0].source_list[0] = source

        # Change the source hypo group manager of the EventSelectionMethod
        # instance.
        for event_selection_method in self._event_selection_method_list:
            if(event_selection_method is not None):
                event_selection_method.change_source_hypo_group_manager(
                    self._src_hypo_group_manager)

        # Change the source hypo group manager of the LLH ratio function
        # instance.
        self._llhratio.change_source_hypo_group_manager(self._src_hypo_group_manager)

        # Change the source hypo group manager of the background generator
        # instance.
        if(self._bkg_generator is not None):
            self._bkg_generator.change_source_hypo_group_manager(
                self._src_hypo_group_manager)

        # Change the source hypo group manager of the signal generator instance.
        if(self._sig_generator is not None):
            self._sig_generator.change_source_hypo_group_manager(
                self._src_hypo_group_manager)

    def change_sources(self, sources):
        """Changes the sources of the analysis to the given source list. It
        makes the necessary changes to all the objects of the analysis.

        Parameters
        ----------
        sources : list of SourceModel instances
            The SourceModel instances describing new sources.
        """
        if(isinstance(sources, SourceModel)):
            sources = [sources]
        if(not issequenceof(sources, SourceModel)):
            raise TypeError('The sources argument must be a list of instances '
                            'of SourceModel')

        if(self._llhratio is None):
            raise RuntimeError(
                'The LLH ratio function has to be constructed, '
                'before the `change_source` method can be called!')

        # Change the source in the SourceHypoGroupManager instance.
        # Because this is a single type sources analysis, there can only be one
        # source hypothesis group defined.
        self._src_hypo_group_manager.src_hypo_group_list[0].source_list = sources

        # Change the source hypo group manager of the EventSelectionMethod
        # instance.
        for event_selection_method in self._event_selection_method_list:
            if(event_selection_method is not None):
                event_selection_method.change_source_hypo_group_manager(
                    self._src_hypo_group_manager)

        # Change the source hypo group manager of the LLH ratio function
        # instance.
        self._llhratio.change_source_hypo_group_manager(
            self._src_hypo_group_manager)

        # Change the source hypo group manager of the background generator
        # instance.
        if(self._bkg_generator is not None):
            self._bkg_generator.change_source_hypo_group_manager(
                self._src_hypo_group_manager)

        # Change the source hypo group manager of the signal generator instance.
        if(self._sig_generator is not None):
            self._sig_generator.change_source_hypo_group_manager(
                self._src_hypo_group_manager)

    def initialize_trial(self, events_list, n_events_list=None, tl=None):
        """This method initializes the multi-dataset log-likelihood ratio
        function with a new set of given trial data. This is a low-level method.
        For convenient methods see the `unblind` and `do_trial` methods.

        Parameters
        ----------
        events_list : list of DataFieldRecordArray instances
            The list of DataFieldRecordArray instances holding the data events
            to use for the log-likelihood function evaluation. The data arrays
            for the datasets must be in the same order than the added datasets.
        n_events_list : list of int | None
            The list of the number of events of each data set. If set to None,
            the number of events is taken from the size of the given events
            arrays.
        tl : TimeLord | None
            The optional TimeLord instance that should be used for timing
            measurements.
        """
        if(n_events_list is None):
            n_events_list = [None] * len(events_list)

        for (idx, (tdm, events, n_events, evt_sel_method)) in enumerate(zip(
                self._tdm_list, events_list, n_events_list,
                self._event_selection_method_list)):

            # Initialize the trial data manager with the given raw events.
            self._tdm_list[idx].initialize_trial(
                self._src_hypo_group_manager, events, n_events, evt_sel_method,
                tl=tl)

        self._llhratio.initialize_for_new_trial(tl=tl)

    def maximize_llhratio(self, rss, tl=None):
        """Maximizes the log-likelihood ratio function, by minimizing its
        negative.

        Parameters
        ----------
        rss : RandomStateService instance
            The RandomStateService instance that should be used to draw random
            numbers from. It is used by the minimizer to generate random
            fit parameter initial values.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to time the
            maximization of the LLH ratio function.

        Returns
        -------
        fitparamset : FitParameterSet instance
            The instance of FitParameterSet holding the global fit parameter
            definitions used in the maximization process.
        log_lambda_max : float
            The value of the log-likelihood ratio function at its maximum.
        fitparam_values : (N_fitparam,)-shaped 1D ndarray
            The ndarray holding the global fit parameter values.
            By definition, the first element is the value of the fit parameter
            ns.
        status : dict
            The dictionary with status information about the maximization
            process, i.e. from the minimizer.
        """
        (log_lambda_max, fitparam_values, status) = self._llhratio.maximize(
            rss, self._fitparamset, tl=tl)
        return (self._fitparamset, log_lambda_max, fitparam_values, status)

    def calculate_fluxmodel_scaling_factor(self, mean_ns, fitparam_values):
        """Calculates the factor the source's fluxmodel has to be scaled
        in order to obtain the given mean number of signal events in the
        detector.

        Parameters
        ----------
        mean_ns : float
            The mean number of signal events in the detector for which the
            scaling factor is calculated.
        fitparam_values : (N_fitparams,)-shaped 1D ndarray
            The ndarray holding the fit parameter values that should be used for
            the flux calculation.

        Returns
        -------
        factor : float
            The factor the given fluxmodel needs to be scaled in order to obtain
            the given mean number of signal events in the detector.
        """
        fitparams_arr = self._src_fitparam_mapper.get_fitparams_array(
            fitparam_values)

        # We use the DatasetSignalWeights class instance of this analysis to
        # calculate the detector signal yield for all datasets.
        dataset_signal_weights = self._llhratio.dataset_signal_weights

        # Calculate the detector signal yield, i.e. the mean number of signal
        # events in the detector, for the given reference flux model.
        mean_ns_ref = 0
        detsigyields = dataset_signal_weights.detsigyield_arr[0]
        for detsigyield in detsigyields:
            (Yj, Yj_grads) = detsigyield(
                dataset_signal_weights._src_arr_list[0], fitparams_arr)
            mean_ns_ref += Yj[0]

        factor = mean_ns / mean_ns_ref

        return factor


class TimeIntegratedMultiDatasetMultiSourceAnalysis(
        TimeIntegratedMultiDatasetSingleSourceAnalysis):
    """This is an analysis class that implements a time-integrated LLH ratio
    analysis for multiple datasets assuming multiple sources.

    To run this analysis the following procedure applies:

        1. Add the datasets and their spatial and energy PDF ratio instances
           via the :meth:`.add_dataset` method.
        2. Construct the log-likelihood ratio function via the
           :meth:`construct_llhratio` method.
        3. Initialize a trial via the :meth:`initialize_trial` method.
        4. Fit the global fit parameters to the trial data via the
           :meth:`maximize_llhratio` method.
    """
    def __init__(
            self, src_hypo_group_manager, src_fitparam_mapper, fitparam_ns,
            test_statistic, bkg_gen_method=None, sig_generator_cls=None):
        """Creates a new time-integrated point-like source analysis assuming
        multiple sources.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses, their flux model, and their detector signal
            efficiency implementation method.
        src_fitparam_mapper : instance of SingleSourceFitParameterMapper
            The instance of SingleSourceFitParameterMapper defining the global
            fit parameters and their mapping to the source fit parameters.
        fitparam_ns : FitParameter instance
            The FitParameter instance defining the fit parameter ns.
        test_statistic : TestStatistic instance
            The TestStatistic instance that defines the test statistic function
            of the analysis.
        bkg_gen_method : instance of BackgroundGenerationMethod | None
            The instance of BackgroundGenerationMethod that will be used to
            generate background events for a new analysis trial. This can be set
            to None, if no background events have to get generated.
        sig_generator_cls : SignalGeneratorBase class | None
            The signal generator class used to create the signal generator
            instance.
            If set to None, the `SignalGenerator` class is used.
        """
        super().__init__(
            src_hypo_group_manager=src_hypo_group_manager,
            src_fitparam_mapper=src_fitparam_mapper,
            fitparam_ns=fitparam_ns,
            test_statistic=test_statistic,
            bkg_gen_method=bkg_gen_method,
            sig_generator_cls=sig_generator_cls)

    def construct_llhratio(self, minimizer, ppbar=None):
        """Constructs the log-likelihood-ratio (LLH-ratio) function of the
        analysis. This setups all the necessary analysis
        objects like detector signal efficiencies and dataset signal weights,
        constructs the log-likelihood ratio functions for each dataset and the
        final composite llh ratio function.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The instance of Minimizer that should be used to minimize the
            negative of the log-likelihood ratio function.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        llhratio : instance of MultiDatasetTCLLHRatio
            The instance of MultiDatasetTCLLHRatio that implements the
            log-likelihood-ratio function of the analysis.
        """
        # Create the detector signal yield instances for each dataset.
        # Multi source analysis has to also support multiple source hypothesis
        # groups.
        # Initialize empty (N_source_hypo_groups, N_datasets)-shaped ndarray.
        detsigyield_array = np.empty(
            (self._src_hypo_group_manager.n_src_hypo_groups,
             len(self.dataset_list)),
            dtype=object
        )

        for (g, shg) in enumerate(self._src_hypo_group_manager._src_hypo_group_list):
            fluxmodel = shg.fluxmodel
            detsigyield_implmethod_list = shg.detsigyield_implmethod_list

            if((len(detsigyield_implmethod_list) != 1) and
               (len(detsigyield_implmethod_list) != self.n_datasets)):
                raise ValueError(
                    'The number of detector signal yield '
                    'implementation methods is not 1 and does not match the number '
                    'of used datasets in the analysis!')
            pbar = ProgressBar(len(self.dataset_list), parent=ppbar).start()
            for (j, (dataset, data)) in enumerate(zip(self.dataset_list,
                                                      self.data_list)):
                if(len(detsigyield_implmethod_list) == 1):
                    # Only one detsigyield implementation method was defined, so we
                    # use it for all datasets.
                    detsigyield_implmethod = detsigyield_implmethod_list[0]
                else:
                    detsigyield_implmethod = detsigyield_implmethod_list[j]

                detsigyield = detsigyield_implmethod.construct_detsigyield(
                    dataset, data, fluxmodel, data.livetime, ppbar=pbar)
                detsigyield_array[g, j] = detsigyield
                pbar.increment()
            pbar.finish()

        # For multiple datasets we need a dataset signal weights instance in
        # order to distribute ns over the different datasets.
        dataset_signal_weights = MultiSourceDatasetSignalWeights(
            self._src_hypo_group_manager, self._src_fitparam_mapper,
            detsigyield_array)

        # Create the list of log-likelihood ratio functions, one for each
        # dataset.
        llhratio_list = []
        for j in range(self.n_datasets):
            tdm = self._tdm_list[j]
            pdfratio_list = self._pdfratio_list_list[j]
            llhratio = MultiSourceZeroSigH0SingleDatasetTCLLHRatio(
                minimizer,
                self._src_hypo_group_manager,
                self._src_fitparam_mapper,
                tdm,
                pdfratio_list,
                detsigyield_array[:, j]
            )
            llhratio_list.append(llhratio)

        # Create the final multi-dataset log-likelihood ratio function.
        llhratio = MultiDatasetTCLLHRatio(
            minimizer, dataset_signal_weights, llhratio_list)

        return llhratio

    def initialize_trial(self, events_list, n_events_list=None, tl=None):
        """This method initializes the multi-dataset log-likelihood ratio
        function with a new set of given trial data. This is a low-level method.
        For convenient methods see the `unblind` and `do_trial` methods.

        Parameters
        ----------
        events_list : list of DataFieldRecordArray instances
            The list of DataFieldRecordArray instances holding the data events
            to use for the log-likelihood function evaluation. The data arrays
            for the datasets must be in the same order than the added datasets.
        n_events_list : list of int | None
            The list of the number of events of each data set. If set to None,
            the number of events is taken from the size of the given events
            arrays.
        tl : TimeLord | None
            The optional TimeLord instance that should be used for timing
            measurements.
        """
        if(n_events_list is None):
            n_events_list = [None] * len(events_list)

        for (idx, (tdm, events, n_events, evt_sel_method)) in enumerate(zip(
                self._tdm_list, events_list, n_events_list,
                self._event_selection_method_list)):

            # Initialize the trial data manager with the given raw events.
            self._tdm_list[idx].initialize_trial(
                self._src_hypo_group_manager, events, n_events, evt_sel_method,
                store_src_ev_idxs=True, tl=tl)

        self._llhratio.initialize_for_new_trial(tl=tl)



# -*- coding: utf-8 -*-

"""Plotting module for core PDF ratio objects.
"""

import numpy as np
import itertools

from matplotlib.axes import Axes
from matplotlib.colors import LogNorm

from skyllh.core.py import classname
from skyllh.core.source_hypothesis import SourceHypoGroupManager
from skyllh.core.storage import DataFieldRecordArray
from skyllh.core.trialdata import TrialDataManager
from skyllh.core.pdfratio import SpatialSigOverBkgPDFRatio


class SpatialSigOverBkgPDFRatioPlotter(object):
    """Plotter class to plot a SpatialSigOverBkgPDFRatio object.
    """
    def __init__(self, tdm, pdfratio):
        """Creates a new plotter object for plotting a
        SpatialSigOverBkgPDFRatio object.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that provides the data for the
            PDF ratio evaluation.
        pdfratio : SpatialSigOverBkgPDFRatio
            The PDF ratio object to plot.
        """
        self.tdm = tdm
        self.pdfratio = pdfratio

    @property
    def pdfratio(self):
        """The PDF ratio object to plot.
        """
        return self._pdfratio
    @pdfratio.setter
    def pdfratio(self, pdfratio):
        if(not isinstance(pdfratio, SpatialSigOverBkgPDFRatio)):
            raise TypeError('The pdfratio property must be an object of instance SpatialSigOverBkgPDFRatio!')
        self._pdfratio = pdfratio

    @property
    def tdm(self):
        """The TrialDataManager that provides the data for the PDF evaluation.
        """
        return self._tdm
    @tdm.setter
    def tdm(self, obj):
        if(not isinstance(obj, TrialDataManager)):
            raise TypeError('The tdm property must be an instance of '
                'TrialDataManager!')
        self._tdm = obj

    def plot(self, src_hypo_group_manager, axes, source_idx=None, log=True,
            **kwargs):
        """Plots the spatial PDF ratio. If the signal PDF depends on the source,
        source_idx specifies the index of the source for which the PDF should
        get plotted.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the source
            hypotheses.
        axes : mpl.axes.Axes
            The matplotlib Axes object on which the PDF ratio should get drawn
            to.
        source_idx : int | None
            The index of the source for which the PDF ratio should get plotted.
            If set to None and the signal PDF depends on the source, index 0
            will be used.

        Additional Keyword Arguments
        ----------------------------
        Any additional keyword arguments will be passed to the `mpl.imshow`
        function.

        Returns
        -------
        img : instance of mpl.AxesImage
            The AxesImage instance showing the PDF ratio image.
        """
        if(not isinstance(axes, Axes)):
            raise TypeError('The axes argument must be an instance of matplotlib.axes.Axes!')

        if(source_idx is None):
            source_idx = 0

        # Define the binning for ra, dec, and sin_dec.
        delta_ra_deg = 0.5
        delta_dec_deg = 0.5

        raaxis = self._pdfratio.signalpdf.axes.get_axis('ra')
        decaxis = self._pdfratio.signalpdf.axes.get_axis('dec')

        # Create a grid of ratio in right-ascention and declination and fill it
        # with PDF ratio values from events that fall into these bins.
        # Use a binning for 1/2 degree.
        rabins = int(np.ceil(raaxis.length / np.deg2rad(delta_ra_deg)))
        decbins = int(np.ceil(decaxis.length / np.deg2rad(delta_dec_deg)))

        ratios = np.zeros((rabins,decbins), dtype=np.float64)

        ra_binedges = np.linspace(raaxis.vmin, raaxis.vmax, rabins+1)
        ra_bincenters = 0.5*(ra_binedges[:-1] + ra_binedges[1:])

        dec_binedges = np.linspace(decaxis.vmin, decaxis.vmax, decbins+1)
        dec_bincenters = 0.5*(dec_binedges[:-1] + dec_binedges[1:])

        # Generate events that fall into the ratio bins.
        events = DataFieldRecordArray(
            np.zeros(
                (ratios.size,),
                dtype=[('ira', np.int64), ('ra', np.float64),
                       ('idec', np.int64), ('dec', np.float64),
                       ('sin_dec', np.float64),
                       ('ang_err', np.float64)]))
        for (i, ((ira,ra),(idec,dec))) in enumerate(itertools.product(
                                                enumerate(ra_bincenters),
                                                enumerate(dec_bincenters))):
            events['ira'][i] = ira
            events['ra'][i] = ra
            events['idec'][i] = idec
            events['dec'][i] = dec
            events['sin_dec'][i] = np.sin(dec)
            events['ang_err'][i] = np.deg2rad(0.5)

        self._tdm.initialize_for_new_trial(src_hypo_group_manager, events)

        event_ratios = self._pdfratio.get_ratio(self._tdm)

        # Select only the ratios for the requested source.
        if(event_ratios.ndim == 2):
            event_ratios = event_ratios[source_idx]

        ratios[events['ira'],events['idec']] = event_ratios

        (left, right, bottom, top) = (raaxis.vmin, raaxis.vmax,
                                      decaxis.vmin, decaxis.vmax)
        norm = LogNorm() if log else None
        img = axes.imshow(ratios.T,
            extent=(left, right, bottom, top),
            origin='lower',
            norm=norm,
            interpolation='none', **kwargs)
        axes.set_xlabel(raaxis.name)
        axes.set_ylabel(decaxis.name)
        axes.set_title(classname(self._pdfratio))

        return img



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.py import classname

class BinningDefinition(object):
    """The BinningDefinition class provides a structure to hold histogram
    binning definitions for an analyis.
    """
    def __init__(self, name, binedges):
        """Creates a new binning definition object.

        Parameters
        ----------
        name : str
            The name of the binning definition.
        binedges : sequence
            The sequence of the bin edges, which should be used for the binning.
        """
        self.name = name
        self.binedges = binedges

    def __str__(self):
        """Pretty string representation.
        """
        s = '%s: %s\n'%(classname(self), self._name)
        s += str(self._binedges)
        return s

    def __eq__(self, other):
        """Checks if object ``other`` is equal to this BinningDefinition object.
        """
        if(not isinstance(other, BinningDefinition)):
            raise TypeError('The other object in the equal comparison must be '
                'an instance of BinningDefinition!')
        if(self.name != other.name):
            return False
        if(np.any(self.binedges != other.binedges)):
            return False
        return True

    @property
    def name(self):
        """The name of the binning setting. This must be an unique name
        for all the different binning settings used within a season.
        """
        return self._name
    @name.setter
    def name(self, name):
        if(not isinstance(name, str)):
            raise TypeError("The name must be of type 'str'!")
        self._name = name

    @property
    def binedges(self):
        """The numpy.ndarray holding the bin edges.
        """
        return self._binedges
    @binedges.setter
    def binedges(self, arr):
        arr = np.atleast_1d(arr)
        self._binedges = np.array(arr, dtype=np.float64)

    @property
    def nbins(self):
        """The number of bins, based on the number of bin edges (minus 1).
        """
        return self._binedges.size - 1

    @property
    def bincenters(self):
        """The center values of the bins.
        """
        return 0.5*(self._binedges[:-1] + self._binedges[1:])

    @property
    def lower_edge(self):
        """The lowest bin edge of the binning.
        """
        return self._binedges[0]

    @property
    def upper_edge(self):
        """The upper most edge of the binning.
        """
        return self._binedges[-1]

    @property
    def range(self):
        """The tuple (lower_edge, upper_edge) of the binning.
        """
        return (self.lower_edge, self.upper_edge)

    def any_data_out_of_binning_range(self, data):
        """Checks if any of the given data is outside of the binning range.

        Parameters
        ----------
        data : 1d ndarray
            The array with the data values to check.

        Returns
        -------
        outofrange : bool
            True if any data value is outside the binning range.
            False otherwise.
        """
        outofrange = np.any((data < self.lower_edge) |
                            (data > self.upper_edge))
        return outofrange

    def get_subset(self, lower_edge, upper_edge):
        """Creates a new BinningDefinition instance which contains only a subset
        of the bins of this BinningDefinition instance. The range of the subset
        is given by a lower and upper edge value.

        Parameters
        ----------
        lower_edge : float
            The lower edge value of the subset.
        upper_edge : float
            The upper edge value of the subset.

        Returns
        -------
        new_binning : BinningDefinition instance
            The new BinningDefinition instance holding the binning subset.
        """

        idxs = np.indices((len(self._binedges),))[0]
        m = (self._binedges >= lower_edge) & (self._binedges <= upper_edge)

        idx_lower = np.min(idxs[m])
        # Include the lower edge of the bin the lower_edge value falls into.
        if(self._binedges[idx_lower] > lower_edge):
            idx_lower -= 1

        idx_upper = np.max(idxs[m])
        # Include the upper edge of the bin the upper_edge value falls into.
        if(self._binedges[idx_upper] < upper_edge):
            idx_upper += 1

        new_binedges = self._binedges[idx_lower:idx_upper+1]

        return BinningDefinition(self._name, new_binedges)


class UsesBinning(object):
    """This is a classifier class that can be used to define, that a class uses
    binning information for one or more dimensions.

    This class defines the property ``binnings``, which is a list of
    BinningDefinition objects.

    This class provides the method ``has_same_binning_as(obj)`` to determine if
    a given object (that also uses binning) has the same binning.
    """
    def __init__(self, *args, **kwargs):
        # Make sure that multiple inheritance can be used.
        super(UsesBinning, self).__init__(*args, **kwargs)

        # Define the list of binning definition objects and a name->list_index
        # mapping for faster access.
        self._binnings = []
        self._binning_name2idx = {}

    @property
    def binnings(self):
        """(read-only) The list of BinningDefinition objects, one for each
        dimension.
        """
        return self._binnings

    @property
    def binning_ndim(self):
        """(read-only) The number of dimensions that uses binning.
        """
        return len(self._binnings)

    def has_same_binning_as(self, obj):
        """Checks if this object has the same binning as the given object.

        Parameters
        ----------
        obj : class instance derived from UsesBinning
            The object that should be checked for same binning.

        Returns
        -------
        check : bool
            True if ``obj`` uses the same binning, False otherwise.
        """
        if(not isinstance(obj, UsesBinning)):
            raise TypeError('The obj argument must be an instance of '
                'UsesBinning!')

        for (self_binning, obj_binning) in zip(self.binnings, obj.binnings):
            if(not (self_binning == obj_binning)):
                return False
        return True

    def add_binning(self, binning, name=None):
        """Adds the given binning definition to the list of binnings.

        Parameters
        ----------
        binning : BinningDefinition
            The binning definition to add.
        name : str | (default) None
            The name of the binning. If not None and it's different to the
            name of the given binning definition, a copy of the
            BinningDefinition object is made and the new name is set.
        """
        if(not isinstance(binning, BinningDefinition)):
            raise TypeError('The binning argument must be an instance of '
                'BinningDefinition!')

        # Create a copy of the BinningDefinition object if the name differs.
        if(name is not None):
            if(not isinstance(name, str)):
                raise TypeError('The name argument must be of type str!')
            if(name != binning.name):
                binning = BinningDefinition(name, binning.binedges)

        self._binnings.append(binning)
        self._binning_name2idx[binning.name] = len(self._binnings)-1

    def get_binning(self, name):
        """Retrieves the binning definition of the given name.

        Parameters
        ----------
        name : str | int
            The name of the binning definition. A string specifies the name and
            an integer the dimension index.

        Returns
        -------
        binning : BinningDefinition
            The binning definition of the given name.
        """
        if(isinstance(name, str)):
            if(name not in self._binning_name2idx):
                raise KeyError('The binning definition "%s" is not defined!'%(
                    name))
            binning = self._binnings[self._binning_name2idx[name]]
        elif(isinstance(name, int)):
            binning = self._binnings[name]
        else:
            raise TypeError('The name argument must be of type str or int!')

        return binning



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.binning import BinningDefinition
from skyllh.core.multiproc import (
    IsParallelizable,
    parallelize
)
from skyllh.core.parameters import (
    ParameterGrid,
    ParameterGridSet
)
from skyllh.core.smoothing import SmoothingFilter
from skyllh.core.pdf import (
    PDFSet,
    IsSignalPDF
)
from skyllh.physics.flux import FluxModel
from skyllh.physics.source import PointLikeSource
from skyllh.i3.pdf import I3EnergyPDF


class SignalI3EnergyPDFSet(PDFSet, IsSignalPDF, IsParallelizable):
    """This is the signal energy PDF for IceCube. It creates a set of
    I3EnergyPDF objects for a discrete set of energy signal parameters. Energy
    signal parameters are the parameters that influence the source flux model.
    """
    def __init__(self, data_mc, logE_binning, sinDec_binning, fluxmodel,
                 fitparam_grid_set, smoothing_filter=None, ncpu=None,
                 ppbar=None):
        """Creates a new IceCube energy signal PDF for a given flux model and
        a set of fit parameter grids for the flux model.
        It creates a set of I3EnergyPDF objects for each signal parameter value
        permutation and stores it inside the ``_params_hash_I3EnergyPDF_dict``
        dictionary, where the hash of the fit parameters dictionary is the key.

        Parameters
        ----------
        data_mc : instance of DataFieldRecordArray
            The array holding the monte-carlo data. The following data fields
            must exist:

            - 'true_energy' : float
                The true energy value of the data event.
            - 'log_energy' : float
                The logarithm of the reconstructed energy value of the data
                event.
            - 'dec' : float
                The declination of the data event.
            - 'mcweight' : float
                The monte-carlo weight value of the data events in unit
                GeV cm^2 sr.

        logE_binning : BinningDefinition
            The binning definition for the binning in log10(E).
        sinDec_binning : BinningDefinition
            The binning definition for the sin(declination).
        fluxmodel : FluxModel
            The flux model to use to create the signal energy PDF.
        fitparam_grid_set : ParameterGridSet | ParameterGrid
            The set of parameter grids. A ParameterGrid object for each
            energy fit parameter, for which an I3EnergyPDF object needs to be
            created.
        smoothing_filter : SmoothingFilter instance | None
            The smoothing filter to use for smoothing the energy histogram.
            If None, no smoothing will be applied.
        ncpu : int | None (default)
            The number of CPUs to use to create the different I3EnergyPDF
            objects for the different fit parameter grid values.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar of the optional parent progress bar.
        """
        if(isinstance(fitparam_grid_set, ParameterGrid)):
            fitparam_grid_set = ParameterGridSet([fitparam_grid_set])
        if(not isinstance(fitparam_grid_set, ParameterGridSet)):
            raise TypeError('The fitparam_grid_set argument must be an '
                'instance of ParameterGrid or ParameterGridSet!')

        # We need to extend the fit parameter grids on the lower and upper end
        # by one bin to allow for the calculation of the interpolation. But we
        # will do this on a copy of the object.
        fitparam_grid_set = fitparam_grid_set.copy()
        fitparam_grid_set.add_extra_lower_and_upper_bin()

        super(SignalI3EnergyPDFSet, self).__init__(pdf_type=I3EnergyPDF,
            fitparams_grid_set=fitparam_grid_set, ncpu=ncpu)

        if(not isinstance(logE_binning, BinningDefinition)):
            raise TypeError('The logE_binning argument must be an instance of '
                'BinningDefinition!')
        if(not isinstance(sinDec_binning, BinningDefinition)):
            raise TypeError('The sinDec_binning argument must be an instance '
                'of BinningDefinition!')
        if(not isinstance(fluxmodel, FluxModel)):
            raise TypeError('The fluxmodel argument must be an instance of '
                'FluxModel!')
        if((smoothing_filter is not None) and
           (not isinstance(smoothing_filter, SmoothingFilter))):
            raise TypeError('The smoothing_filter argument must be None or '
                'an instance of SmoothingFilter!')

        # Create I3EnergyPDF objects for all permutations of the fit parameter
        # grid values.
        def create_I3EnergyPDF(
            data_logE, data_sinDec, data_mcweight, data_true_energy,
            logE_binning, sinDec_binning, smoothing_filter, fluxmodel,
            gridfitparams):
            """Creates an I3EnergyPDF object for the given flux model and flux
            parameters.

            Parameters
            ----------
            data_logE : 1d ndarray
                The logarithm of the reconstructed energy value of the data
                events.
            data_sinDec : 1d ndarray
                The sin(dec) value of the the data events.
            data_mcweight : 1d ndarray
                The monte-carlo weight value of the data events.
            data_true_energy : 1d ndarray
                The true energy value of the data events.
            logE_binning : BinningDefinition
                The binning definition for the binning in log10(E).
            sinDec_binning : BinningDefinition
                The binning definition for the sin(declination).
            smoothing_filter : SmoothingFilter instance | None
                The smoothing filter to use for smoothing the energy histogram.
                If None, no smoothing will be applied.
            fluxmodel : FluxModel
                The flux model to use to create the signal event weights.
            gridfitparams : dict
                The dictionary holding the specific signal flux parameters.

            Returns
            -------
            i3energypdf : I3EnergyPDF
                The created I3EnergyPDF object for the given flux model and flux
                parameters.
            """
            # Create a copy of the FluxModel with the given flux parameters.
            # The copy is needed to not interfer with other CPU processes.
            myfluxmodel = fluxmodel.copy(newprop=gridfitparams)

            # Calculate the signal energy weight of the event. Note, that
            # because we create a normalized PDF, we can ignore all constants.
            # So we don't have to convert the flux unit into the internally used
            # flux unit.
            data_physicsweight = myfluxmodel(data_true_energy)

            i3energypdf = I3EnergyPDF(
                data_logE, data_sinDec, data_mcweight, data_physicsweight,
                logE_binning, sinDec_binning, smoothing_filter)

            return i3energypdf

        data_logE = data_mc['log_energy']
        data_sinDec = np.sin(data_mc['dec'])
        data_mcweight = data_mc['mcweight']
        data_true_energy = data_mc['true_energy']

        args_list = [ ((data_logE, data_sinDec, data_mcweight, data_true_energy,
                        logE_binning, sinDec_binning, smoothing_filter,
                        fluxmodel, gridfitparams), {})
                     for gridfitparams in self.gridfitparams_list ]

        i3energypdf_list = parallelize(
            create_I3EnergyPDF, args_list, self.ncpu, ppbar=ppbar)

        # Save all the I3EnergyPDF objects in the IsSignalPDF PDF registry with
        # the hash of the individual parameters as key.
        for (gridfitparams, i3energypdf) in zip(self.gridfitparams_list, i3energypdf_list):
            self.add_pdf(i3energypdf, gridfitparams)

    def assert_is_valid_for_exp_data(self, data_exp):
        """Checks if this signal energy PDF is valid for all the given
        experimental data.
        It checks if all the data is within the logE and sin(dec) binning range.

        Parameters
        ----------
        data_exp : numpy record ndarray
            The array holding the experimental data. The following data fields
            must exist:

            - 'log_energy' : float
                The logarithm of the energy value of the data event.
            - 'dec' : float
                The declination of the data event.

        Raises
        ------
        ValueError
            If some of the data is outside the logE or sin(dec) binning range.
        """
        # Since we use the same binning for all the I3EnergyPDF objects, we
        # can just use an arbitrary object to verify the data.
        self.get_pdf(self.pdf_keys[0]).assert_is_valid_for_exp_data(data_exp)

    def get_prob(self, tdm, gridfitparams):
        """Calculates the signal energy probability (in logE) of each event for
        a given set of signal fit parameters on a grid.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the data events for which the
            probability should be calculated for. The following data fields must
            exist:

            - 'log_energy' : float
                The logarithm of the energy value of the event.
            - 'sin_dec' : float
                The sin(declination) value of the event.

        gridfitparams : dict
            The dictionary holding the signal parameter values for which the
            signal energy probability should be calculated. Note, that the
            parameter values must match a set of parameter grid values for which
            an I3EnergyPDF object has been created at construction time of this
            SignalI3EnergyPDF object. There is no interpolation method defined
            at this point to allow for arbitrary parameter values!

        Returns
        -------
        prob : 1d ndarray
            The array with the signal energy probability for each event.

        Raises
        ------
        KeyError
            If no energy PDF can be found for the given signal parameter values.
        """
        i3energypdf = self.get_pdf(gridfitparams)

        prob = i3energypdf.get_prob(tdm)
        return prob



# -*- coding: utf-8 -*-

import unittest
import numpy as np
from numpy.testing import assert_array_almost_equal

from skyllh.core.storage import DataFieldRecordArray


class DataFieldRecordArray_TestCase(unittest.TestCase):
    def setUp(self):
        self.field1 = np.array([1.4, 1.3, 1.5, 1.1, 1.2], dtype=np.float64)
        self.field2 = np.array([2.5, 2.1, 2.3, 2.4, 2.2], dtype=np.float64)
        self.field3 = np.array([3.2, 3.5, 3.1, 3.3, 3.4], dtype=np.float64)
        data = dict(
            field1 = self.field1,
            field2 = self.field2,
            field3 = self.field3
        )
        self.arr = DataFieldRecordArray(data)
        self.arr_len = 5

    def test__contains__(self):
        self.assertFalse('field0' in self.arr)
        self.assertTrue('field1' in self.arr)
        self.assertTrue('field2' in self.arr)
        self.assertTrue('field3' in self.arr)

    def test__getitem__(self):
        # Access a non-existing field.
        with self.assertRaises(KeyError):
            self.arr['field0']

        # Access entire fields.
        assert_array_almost_equal(self.arr['field1'], self.field1)
        assert_array_almost_equal(self.arr['field2'], self.field2)
        assert_array_almost_equal(self.arr['field3'], self.field3)

        # Access rows of the dataset via indices.
        idx = np.array([1,4,2])
        sub_arr = self.arr[idx]
        assert_array_almost_equal(sub_arr['field1'], self.field1[idx])
        assert_array_almost_equal(sub_arr['field2'], self.field2[idx])
        assert_array_almost_equal(sub_arr['field3'], self.field3[idx])

        # Access rows of the dataset via a boolean mask.
        mask = np.array([True, True, False, True, False])
        sub_arr = self.arr[mask]
        assert_array_almost_equal(sub_arr['field1'], self.field1[mask])
        assert_array_almost_equal(sub_arr['field2'], self.field2[mask])
        assert_array_almost_equal(sub_arr['field3'], self.field3[mask])

    def test__setitem__(self):
        # Set an entire field with data of not the same length.
        with self.assertRaises(ValueError):
            new_field2 = np.array([2.51, 2.12, 2.33, 2.44], dtype=np.float64)
            self.arr['field2'] = new_field2

        # Set an entire field with new values.
        new_field2 = np.array([2.51, 2.12, 2.33, 2.44, 2.25], dtype=np.float64)
        self.arr['field2'] = new_field2
        assert_array_almost_equal(self.arr['field2'], new_field2)

        # Reset the array.
        self.setUp()

        # Set selected rows with new values by indices.
        idx = np.array([1,4,2])
        new_data = dict(
            field1 = self.field1[idx],
            field2 = new_field2[idx],
            field3 = self.field3[idx]
        )
        new_arr = DataFieldRecordArray(new_data)
        self.arr[idx] = new_arr
        assert_array_almost_equal(self.arr['field1'], self.field1)
        assert_array_almost_equal(
            self.arr['field2'],
            np.array([2.5, 2.12, 2.33, 2.4, 2.25], dtype=np.float64))
        assert_array_almost_equal(self.arr['field3'], self.field3)

        # Reset the array.
        self.setUp()

        # Set selected rows with new values by mask.
        mask = np.array([True, True, False, True, False])
        new_data = dict(
            field1 = self.field1[mask],
            field2 = new_field2[mask],
            field3 = self.field3[mask]
        )
        new_arr = DataFieldRecordArray(new_data)
        self.arr[mask] = new_arr
        assert_array_almost_equal(self.arr['field1'], self.field1)
        assert_array_almost_equal(
            self.arr['field2'],
            np.array([2.51, 2.12, 2.3, 2.44, 2.2], dtype=np.float64))
        assert_array_almost_equal(self.arr['field3'], self.field3)

        # Reset the array.
        self.setUp()

        # Add a new field.
        new_field = np.array([4.2, 4.5, 4.1, 4.3, 4.4], dtype=np.float64)
        self.arr['field4'] = new_field
        self.assertTrue('field4' in self.arr)
        self.assertTrue('field4' in self.arr.field_name_list)
        assert_array_almost_equal(
            self.arr['field4'],
            new_field)

    def test__str__(self):
        try:
            str(self.arr)
        except:
            self.fail('The __str__ method raised an exception!')

    def test_field_name_list(self):
        self.assertEqual(len(self.arr.field_name_list), 3)
        self.assertTrue('field1' in self.arr.field_name_list)
        self.assertTrue('field2' in self.arr.field_name_list)
        self.assertTrue('field3' in self.arr.field_name_list)

    def test_indices(self):
        assert_array_almost_equal(
            self.arr.indices,
            np.array([0, 1, 2, 3, 4]))

    def test_len(self):
        self.assertEqual(len(self.arr), self.arr_len)

    def test_rename_fields(self):
        self.arr.rename_fields({'field2': 'new_field2'})
        self.assertTrue('field1' in self.arr)
        self.assertFalse('field2' in self.arr)
        self.assertTrue('new_field2' in self.arr)
        self.assertTrue('field3' in self.arr)
        self.assertEqual(len(self.arr.field_name_list), 3)
        self.assertTrue('field1' in self.arr.field_name_list)
        self.assertTrue('new_field2' in self.arr.field_name_list)
        self.assertTrue('field3' in self.arr.field_name_list)

    def test_sort_by_field(self):
        self.arr.sort_by_field('field2')
        assert_array_almost_equal(
            self.arr['field1'],
            np.array([1.3, 1.2, 1.5, 1.1, 1.4], dtype=np.float64))
        assert_array_almost_equal(
            self.arr['field2'],
            np.array([2.1, 2.2, 2.3, 2.4, 2.5], dtype=np.float64))
        assert_array_almost_equal(
            self.arr['field3'],
            np.array([3.5, 3.4, 3.1, 3.3, 3.2], dtype=np.float64))

    def test_tidy_up(self):
        self.arr.tidy_up('field2')
        self.assertEqual(len(self.arr.field_name_list), 1)
        self.assertTrue('field2' in self.arr.field_name_list)
        self.assertFalse('field1' in self.arr)
        self.assertTrue('field2' in self.arr)
        self.assertFalse('field3' in self.arr)

        # Reset the array.
        self.setUp()

        self.arr.tidy_up(('field2','field3'))
        self.assertEqual(len(self.arr.field_name_list), 2)
        self.assertTrue('field2' in self.arr.field_name_list)
        self.assertTrue('field3' in self.arr.field_name_list)
        self.assertFalse('field1' in self.arr)
        self.assertTrue('field2' in self.arr)
        self.assertTrue('field3' in self.arr)

if(__name__ == '__main__'):
    unittest.main()



# -*- coding: utf-8 -*-

"""This file defines IceCube specific global configuration.
"""

# Import everything from the core config, So the user sees the same
# functionality as if using the core config.
from skyllh.core.config import *

# Add default analysis required data fields for experimental and monte-carlo
# data that are IceCube specific.
CFG['dataset']['analysis_required_exp_field_names'] += ['azi', 'zen', 'sin_dec']
CFG['dataset']['analysis_required_mc_field_names'] += ['sin_true_dec']



# -*- coding: utf-8 -*-
# TODO: File out of date, tell flake8 to ignore
# flake8: noqa

from __future__ import division

import os.path
import unittest
import numpy as np

from skyllh.core.analysis import Analysis
from skyllh.core.random import RandomStateService

# Classes to define the source hypothesis.
from skyllh.physics.source import PointLikeSource
from skyllh.physics.flux import PowerLawFlux
from skyllh.core.source_hypo_group import SourceHypoGroup
from skyllh.core.source_hypothesis import SourceHypoGroupManager

# Classes to define the fit parameters.
from skyllh.core.parameters import (
    SingleSourceFitParameterMapper,
    FitParameter
)

# Classes for the minimizer.
from skyllh.core.minimizer import Minimizer, LBFGSMinimizerImpl

# Classes for defining the analysis.
from skyllh.core.test_statistic import TestStatisticWilks
#from skyllh.core.analysis import (
#    SpacialEnergyTimeIntegratedMultiDatasetSingleSourceAnalysis as Analysis
#)

"""
class TestAnalysis(unittest.TestCase):
    def setUp(self):
        # path = os.path.abspath(os.path.dirname(__file__))
        # self.exp_data = np.load(os.path.join(path, 'testdata/exp_testdata.npy'))
        # self.mc_data = np.load(os.path.join(path, 'testdata/mc_testdata.npy'))
        # self.livetime_data = np.load(os.path.join(path, 'testdata/livetime_testdata.npy'))

        # Create the minimizer instance.
        minimizer = Minimizer(LBFGSMinimizerImpl())

        # Create a source hypothesis group manager.
        src_hypo_group_manager = SourceHypoGroupManager(SourceHypoGroup(
            source, fluxmodel, detsigeff_implmethod, sig_gen_method))

        # Create a source fit parameter mapper and define the fit parameters.
        src_fitparam_mapper = SingleSourceFitParameterMapper(rss)
        src_fitparam_mapper.def_fit_parameter(fitparam_gamma)
        # Define the test statistic.
        test_statistic = TestStatisticWilks()

        self.analysis = Analysis()
        self.rss = RandomStateService(seed=0)

        # Define the data scrambler with its data scrambling method, which is used
        # for background generation.
        data_scrambler = DataScrambler(UniformRAScramblingMethod(),
                                       inplace_scrambling=True)

        # Create background generation method.
        bkg_gen_method = FixedScrambledExpDataI3BkgGenMethod(data_scrambler)

    def tearDown(self):
        # self.exp_data.close()
        # self.mc_data.close()
        # self.livetime_data.close()
        pass

    def test_do_trials(self):
        N = 10
        ncpu = None

        self.analysis.do_trials(N, self.rss, ncpu=ncpu)
"""

if(__name__ == '__main__'):
    unittest.main()



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.py import int_cast

class RandomStateService(object):
    """The RandomStateService class provides a container for a
    numpy.random.RandomState object, initialized with a given seed. This service
    can then be passed to any function or method within skyllh that requires a
    random number generator.
    """
    def __init__(self, seed=None):
        """Creates a new random state service. The ``random`` property can then
        be used to draw random numbers.

        Parameters
        ----------
        seed : int | None
            The seed to use. If None, the random number generator will be seeded
            randomly. See the numpy documentation for numpy.random.RandomState
            what that means.
        """
        self._seed = int_cast(seed, 'The seed argument must be None, or '
            'castable to type int!', allow_None=True)
        self.random = np.random.RandomState(self._seed)

    @property
    def seed(self):
        """(read-only) The seed (int) of the random number generator.
        None, if not set. To change the seed, use the `reseed` method.
        """
        return self._seed

    @property
    def random(self):
        """The numpy.random.RandomState object.
        """
        return self._random
    @random.setter
    def random(self, random):
        if(not isinstance(random, np.random.RandomState)):
            raise TypeError('The random property must be of type numpy.random.RandomState!')
        self._random = random

    def reseed(self, seed):
        """Reseeds the random number generator with the given seed.

        Parameters
        ----------
        seed : int | None
            The seed to use. If None, the random number generator will be seeded
            randomly. See the numpy documentation for numpy.random.RandomState
            what that means.
        """
        self._seed = int_cast(seed, 'The seed argument must be None or '
            'castable to type int!', allow_None=True)
        self.random.seed(self._seed)



# -*- coding: utf-8 -*-

"""The trialdata module of SkyLLH provides a trial data manager class that
manages the data of an analysis trial. It provides also possible additional data
fields and their calculation, which are required by the particular analysis.
The rational behind this manager is to compute data fields only once, which can
then be used by different analysis objects, like PDF objects.
"""

from collections import OrderedDict
import numpy as np

from skyllh.core.debugging import get_logger
from skyllh.core import display as dsp
from skyllh.core.py import (
    classname,
    func_has_n_args,
    int_cast,
    issequenceof,
    typename
)
from skyllh.core.storage import DataFieldRecordArray


logger = get_logger(__name__)


class DataField(object):
    """This class defines a data field and its calculation that is used by an
    Analysis class instance. The calculation is defined through an external
    function.
    """
    def __init__(
            self, name, func, fitparam_names=None, dt=None):
        """Creates a new instance of DataField that might depend on fit
        parameters.

        Parameters
        ----------
        name : str
            The name of the data field. It serves as the identifier for the
            data field.
        func : callable
            The function that calculates the values of this data field. The call
            signature must be
            `__call__(tdm, src_hypo_group_manager, fitparams)`,
            where `tdm` is the TrialDataManager instance holding the event data,
            `src_hypo_group_manager` is the SourceHypoGroupManager instance, and
            `fitparams` is the dictionary with the current fit parameter names
            and values. If the data field depends solely on source parameters,
            the call signature must be `__call__(tdm, src_hypo_group_manager)`
            instead.
        fitparam_names : sequence of str | None
            The sequence of str instances specifying the names of the fit
            parameters this data field depends on. If set to None, the data
            field does not depend on any fit parameters.
        dt : numpy dtype | str | None
            If specified it defines the data type this data field should have.
            If a str instance is given, it defines the name of the data field
            whose data type should be taken for this data field.
        """
        super(DataField, self).__init__()

        self.name = name
        self.func = func

        if(fitparam_names is None):
            fitparam_names = []
        if(not issequenceof(fitparam_names, str)):
            raise TypeError('The fitparam_names argument must be None or a '
                'sequence of str instances!')
        self._fitparam_name_list = list(fitparam_names)

        self.dt = dt

        # Define the list of fit parameter values for which the fit parameter
        # depend data field values have been calculated for.
        self._fitparam_value_list = [None]*len(self._fitparam_name_list)

        # Define the member variable that holds the numpy ndarray with the data
        # field values.
        self._values = None

        # Define the most efficient `calculate` method for this kind of data
        # field.
        if(func_has_n_args(self._func, 2)):
            self.calculate = self._calc_source_values
        elif(len(self._fitparam_name_list) == 0):
            self.calculate = self._calc_static_values
        else:
            self.calculate = self._calc_fitparam_dependent_values

    @property
    def name(self):
        """The name of the data field.
        """
        return self._name
    @name.setter
    def name(self, name):
        if(not isinstance(name, str)):
            raise TypeError('The name property must be an instance of str!')
        self._name = name

    @property
    def func(self):
        """The function that calculates the data field values.
        """
        return self._func
    @func.setter
    def func(self, f):
        if(not callable(f)):
            raise TypeError('The func property must be a callable object!')
        if((not func_has_n_args(f, 2)) and
           (not func_has_n_args(f, 3))):
            raise TypeError('The func property must be a function with 2 or 3 '
                'arguments!')
        self._func = f

    @property
    def dt(self):
        """The numpy dtype object defining the data type of this data field.
        A str instance defines the name of the data field whose data type should
        be taken for this data field.
        It is None, if there is no explicit data type defined for this data
        field.
        """
        return self._dt
    @dt.setter
    def dt(self, obj):
        if(obj is not None):
            if((not isinstance(obj, np.dtype)) and
               (not isinstance(obj, str))):
                raise TypeError(
                    'The dt property must be None, an instance of numpy.dtype, '
                    'or an instance of str! Currently it is of type %s.'%(
                        str(type(obj))))
        self._dt = obj

    @property
    def values(self):
        """(read-only) The calculated data values of the data field.
        """
        return self._values

    def __str__(self):
        """Pretty string representation of this DataField instance.
        """
        dtype = 'None'
        vmin = np.nan
        vmax = np.nan
        if(self._values is not None):
            dtype = str(self._values.dtype)
            try:
                vmin = np.min(self._values)
            except:
                pass
            try:
                vmax = np.max(self._values)
            except:
                pass
        s = '{}: {}: '.format(classname(self), self.name)
        s +='{dtype: '
        s += '{}, vmin: {: .3e}, vmax: {: .3e}'.format(
            dtype, vmin, vmax)
        s += '}'

        return s

    def _get_desired_dtype(self, tdm):
        """Retrieves the data type this field should have. It's None, if no
        data type was defined for this data field.
        """
        if(self._dt is not None):
            if(isinstance(self._dt, str)):
                self._dt = tdm.get_dtype(self._dt)
        return self._dt

    def _convert_to_desired_dtype(self, tdm, values):
        """Converts the data type of the given values array to the given data
        type.
        """
        dt = self._get_desired_dtype(tdm)
        if(dt is not None):
            values = values.astype(dt, copy=False)
        return values

    def _calc_source_values(
            self, tdm, src_hypo_group_manager, fitparams):
        """Calculates the data field values utilizing the defined external
        function. The data field values solely depend on fixed source
        parameters.
        """
        self._values = self._func(tdm, src_hypo_group_manager)
        if(not isinstance(self._values, np.ndarray)):
            raise TypeError(
                'The calculation function for the data field "%s" must '
                'return an instance of numpy.ndarray! '
                'Currently it is of type "%s".'%(
                    self._name, typename(type(self._values))))

        # Convert the data type.
        self._values = self._convert_to_desired_dtype(tdm, self._values)

    def _calc_static_values(
            self, tdm, src_hypo_group_manager, fitparams):
        """Calculates the data field values utilizing the defined external
        function, that are static and only depend on source parameters.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance this data field is part of and is
            holding the event data.
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        fitparams : dict
            The dictionary holding the current fit parameter names and values.
            By definition this dictionary is empty.
        """
        values = self._func(tdm, src_hypo_group_manager, fitparams)
        if(not isinstance(values, np.ndarray)):
            raise TypeError(
                'The calculation function for the data field "%s" must '
                'return an instance of numpy.ndarray! '
                'Currently it is of type "%s".'%(
                    self._name, typename(type(values))))

        # Convert the data type.
        values = self._convert_to_desired_dtype(tdm, values)

        # Set the data values. This will add the data field to the
        # DataFieldRecordArray if it does not exist yet.
        tdm.events[self._name] = values

    def _calc_fitparam_dependent_values(
            self, tdm, src_hypo_group_manager, fitparams):
        """Calculate data field values utilizing the defined external
        function, that depend on fit parameter values. We check if the fit
        parameter values have changed.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance this data field is part of and is
            holding the event data.
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        fitparams : dict
            The dictionary holding the current fit parameter names and values.
        """
        if(self._name not in tdm.events):
            # It's the first time this method is called, so we need to calculate
            # the data field values for sure.
            values = self._func(tdm, src_hypo_group_manager, fitparams)
            if(not isinstance(values, np.ndarray)):
                raise TypeError(
                    'The calculation function for the data field "%s" must '
                    'return an instance of numpy.ndarray! '
                    'Currently it is of type "%s".'%(
                        self._name, typename(type(values))))

            # Convert the data type.
            values = self._convert_to_desired_dtype(tdm, values)

            # Set the data values. This will add the data field to the
            # DataFieldRecordArray if it does not exist yet.
            tdm.events[self._name] = values

            # We store the fit parameter values for which the field values were
            # calculated for. So they have to get recalculated only when the
            # fit parameter values the field depends on change.
            self._fitparam_value_list = [
                fitparams[name] for name in self._fitparam_name_list
            ]

            return

        for (idx, fitparam_name) in enumerate(self._fitparam_name_list):
            if(fitparams[fitparam_name] != self._fitparam_value_list[idx]):
                # This current fit parameter value has changed. So we need to
                # re-calculate the data field values.
                values = self._func(tdm, src_hypo_group_manager, fitparams)

                # Convert the data type.
                values = self._convert_to_desired_dtype(tdm, values)

                # Set the data values.
                tdm.events[self._name] = values

                # Store the new fit parameter values.
                self._fitparam_value_list = [
                    fitparams[name] for name in self._fitparam_name_list
                ]

                break


class TrialDataManager(object):
    """The TrialDataManager class manages the event data for an analysis trial.
    It provides possible additional data fields and their calculation.
    New data fields can be defined via the `add_data_field` method.
    Whenever a new trial is being initialized the data fields get re-calculated.
    The data trial manager is provided to the PDF evaluation method.
    Hence, data fields are calculated only once.
    """
    def __init__(self, index_field_name=None):
        """Creates a new TrialDataManager instance.

        Parameters
        ----------
        index_field_name : str | None
            The name of the field that should be used as primary index field.
            If provided, the events will be sorted along this data field. This
            might be useful for run-time performance.
        """
        super(TrialDataManager, self).__init__()

        self.index_field_name = index_field_name

        # Define the list of data fields that depend only on the source
        # parameters.
        self._source_data_fields_dict = OrderedDict()

        # Define the list of data fields that are static and should be
        # calculated prior to a possible event selection.
        self._pre_evt_sel_static_data_fields_dict = OrderedDict()

        # Define the list of data fields that are static, i.e. don't depend on
        # any fit parameters. These fields have to be calculated only once when
        # a new evaluation data is available.
        self._static_data_fields_dict = OrderedDict()

        # Define the list of data fields that depend on fit parameters. These
        # data fields have to be re-calculated whenever a fit parameter value
        # changes.
        self._fitparam_data_fields_dict = OrderedDict()

        # Define the member variable that will hold the raw events for which the
        # data fields get calculated.
        self._events = None

        # Define the member variable that holds the source to event index
        # mapping.
        self._src_ev_idxs = None

        # We store an integer number for the trial data state and increase it
        # whenever the state of the trial data changed. This way other code,
        # e.g. PDFs, can determine when the data changed and internal caches
        # must be flushed.
        self._trial_data_state_id = -1

    @property
    def index_field_name(self):
        """The name of the primary index data field. If not None, events will
        be sorted by this data field.
        """
        return self._index_field_name
    @index_field_name.setter
    def index_field_name(self, name):
        if(name is not None):
            if(not isinstance(name, str)):
                raise TypeError(
                    'The index_field_name property must be an instance of '
                    'type str!')
        self._index_field_name = name

    @property
    def events(self):
        """The DataFieldRecordArray instance holding the data events, which
        should get evaluated.
        """
        return self._events
    @events.setter
    def events(self, arr):
        if(not isinstance(arr, DataFieldRecordArray)):
            raise TypeError(
                'The events property must be an instance of '
                'DataFieldRecordArray!')
        self._events = arr

    @property
    def n_events(self):
        """The total number of events of the dataset this trial data manager
        corresponds to.
        """
        return self._n_events
    @n_events.setter
    def n_events(self, n):
        self._n_events = int_cast(
            n, 'The n_events property must be castable to type int!')

    @property
    def n_selected_events(self):
        """(read-only) The number of selected events which should get evaluated.
        """
        return len(self._events)

    @property
    def n_pure_bkg_events(self):
        """(read-only) The number of pure background events, which are not part
        of the trial data, but must be considered for the test-statistic value.
        It is the difference of n_events and n_selected_events.
        """
        return self._n_events - len(self._events)

    @property
    def src_ev_idxs(self):
        """(read-only) The 2-tuple holding the source index and event index
        1d ndarray arrays.
        """
        return self._src_ev_idxs

    @property
    def trial_data_state_id(self):
        """(read-only) The integer ID number of the trial data. This ID number
        can be used to determine when the trial data has changed its state.
        """
        return self._trial_data_state_id

    def __contains__(self, name):
        """Checks if the given data field is defined in this data field manager.

        Parameters
        ----------
        name : str
            The name of the data field.

        Returns
        -------
        check : bool
            True if the data field is defined in this data field manager,
            False otherwise.
        """
        # Check if the data field is part of the original trial data.
        if((self._events is not None) and
           (name in self._events.field_name_list)):
            return True

        # Check if the data field is a user defined data field.
        if((name in self._source_data_fields_dict) or
           (name in self._pre_evt_sel_static_data_fields_dict) or
           (name in self._static_data_fields_dict) or
           (name in self._fitparam_data_fields_dict)):
            return True

        return False

    def __str__(self):
        """Implements pretty string representation of this TrialDataManager
        instance.
        """
        s = classname(self)+':\n'
        s1 = 'Base data fields:\n'
        s2 = str(self._events)
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)
        s += '\n'

        s1 = 'Source data fields:\n'
        s2 = ''
        for (idx, dfield) in enumerate(self._source_data_fields_dict):
            if(idx > 0):
                s2 += '\n'
            s2 += str(dfield)
        if(s2 == ''):
            s2 = 'None'
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)
        s += '\n'

        s1 = 'Pre-event-selection static data fields:\n'
        s2 = ''
        for (idx, dfield) in enumerate(self._pre_evt_sel_static_data_fields_dict):
            if(idx > 0):
                s2 += '\n'
            s2 += str(dfield)
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)
        s += '\n'

        s1 = 'Static data fields:\n'
        s2 = ''
        for (idx, dfield) in enumerate(self._static_data_fields_dict):
            if(idx > 0):
                s2 += '\n'
            s2 += str(dfield)
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)
        s += '\n'

        s1 = 'Fitparam data fields:\n'
        s2 = ''
        for (idx, dfield) in enumerate(self._fitparam_data_fields_dict):
            if(idx > 0):
                s2 += '\n'
            s2 += str(dfield)
        if(s2 == ''):
            s2 = 'None'
        s1 += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s2)
        s += dsp.add_leading_text_line_padding(dsp.INDENTATION_WIDTH, s1)

        return s

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Recalculate the source data fields.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The SourceHypoGroupManager manager that defines the groups of
            source hypotheses.
        """
        self.calculate_source_data_fields(src_hypo_group_manager)

    def initialize_trial(
            self, src_hypo_group_manager, events, n_events=None,
            evt_sel_method=None, store_src_ev_idxs=False, tl=None):
        """Initializes the trial data manager for a new trial. It sets the raw
        events, calculates pre-event-selection data fields, performs a possible
        event selection and calculates the static data fields for the left-over
        events.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of SourceHypoGroupManager that defines the source
            hypothesis groups.
        events : DataFieldRecordArray instance
            The DataFieldRecordArray instance holding the entire raw events.
        n_events : int | None
            The total number of events of the data set this trial data manager
            corresponds to.
            If None, the number of events is taken from the number of events
            present in the ``events`` array.
        evt_sel_method : EventSelectionMethod | None
            The optional event selection method that should be used to select
            potential signal events.
        store_src_ev_idxs : bool
            If the evt_sel_method is not None, it determines if source and
            event indices of the selected events should get calculated and
            stored.
        tl : TimeLord | None
            The optional TimeLord instance that should be used for timing
            measurements.
        """
        # Set the events property, so that the calculation functions of the data
        # fields can access them.
        self.events = events

        if(n_events is None):
            n_events = len(self._events)
        self.n_events = n_events

        # Calculate pre-event-selection data fields that are required by the
        # event selection method.
        self.calculate_pre_evt_sel_static_data_fields(src_hypo_group_manager)

        if(evt_sel_method is not None):
            logger.debug(
                f'Performing event selection method '
                f'"{classname(evt_sel_method)}".')
            (selected_events, src_ev_idxs) = evt_sel_method.select_events(
                self._events, tl=tl, ret_src_ev_idxs=store_src_ev_idxs)
            logger.debug(
                f'Selected {len(selected_events)} out of {len(self._events)} '
                'events.')
            self.events = selected_events
            self._src_ev_idxs = src_ev_idxs

        # Sort the events by the index field, if a field was provided.
        if(self._index_field_name is not None):
            logger.debug(
                'Sorting events in index field "{}"'.format(
                    self._index_field_name))
            self._events.sort_by_field(self._index_field_name)

        # Now calculate all the static data fields. This will increment the
        # trial data state ID.
        self.calculate_static_data_fields(src_hypo_group_manager)

    def add_source_data_field(self, name, func, dt=None):
        """Adds a new data field to the manager. The data field must depend
        solely on source parameters.

        Parameters
        ----------
        name : str
            The name of the data field. It serves as the identifier for the
            data field.
        func : callable
            The function that calculates the data field values. The call
            signature must be
            `__call__(tdm, src_hypo_group_manager, fitparams)`, where
            `tdm` is the TrialDataManager instance holding the event data,
            `src_hypo_group_manager` is the SourceHypoGroupManager instance,
            and `fitparams` is an unused interface argument.
        dt : numpy dtype | str | None
            If specified it defines the data type this data field should have.
            If a str instance is given, it defines the name of the data field
            whose data type should be taken for the data field.
        """
        if(name in self):
            raise KeyError(
                'The data field "%s" is already defined!'%(name))

        data_field = DataField(name, func, dt=dt)

        self._source_data_fields_dict[name] = data_field

    def add_data_field(
            self, name, func, fitparam_names=None, dt=None, pre_evt_sel=False):
        """Adds a new data field to the manager.

        Parameters
        ----------
        name : str
            The name of the data field. It serves as the identifier for the
            data field.
        func : callable
            The function that calculates the data field values. The call
            signature must be
            `__call__(tdm, src_hypo_group_manager, fitparams)`, where
            `tdm` is the TrialDataManager instance holding the event data,
            `src_hypo_group_manager` is the SourceHypoGroupManager instance,
            and `fitparams` is the dictionary with the current fit parameter
            names and values.
        fitparam_names : sequence of str | None
            The sequence of str instances specifying the names of the fit
            parameters this data field depends on. If set to None, it means that
            the data field does not depend on any fit parameters.
        dt : numpy dtype | str | None
            If specified it defines the data type this data field should have.
            If a str instance is given, it defines the name of the data field
            whose data type should be taken for the data field.
        pre_evt_sel : bool
            Flag if this data field should get calculated before potential
            signal events get selected (True), or afterwards (False).
            Default is False.
        """
        if(name in self):
            raise KeyError(
                'The data field "%s" is already defined!'%(name))

        if(pre_evt_sel and (fitparam_names is not None)):
            raise ValueError(
                f'The pre-event-selection data field "{name}" must not depend '
                 'on fit parameters!')

        data_field = DataField(name, func, fitparam_names, dt=dt)

        if(pre_evt_sel):
            self._pre_evt_sel_static_data_fields_dict[name] = data_field
        elif(fitparam_names is None):
            self._static_data_fields_dict[name] = data_field
        else:
            self._fitparam_data_fields_dict[name] = data_field

    def calculate_source_data_fields(self, src_hypo_group_manager):
        """Calculates the data values of the data fields that solely depend on
        source parameters.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        """
        if(len(self._source_data_fields_dict) == 0):
            return

        fitparams = None
        for (name, dfield) in self._source_data_fields_dict.items():
            dfield.calculate(self, src_hypo_group_manager, fitparams)

        self._trial_data_state_id += 1

    def calculate_pre_evt_sel_static_data_fields(self, src_hypo_group_manager):
        """Calculates the data values of the data fields that should be
        available for the event selection method and do not depend on any fit
        parameters.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        """
        if(len(self._pre_evt_sel_static_data_fields_dict) == 0):
            return

        fitparams = dict()
        for (name, dfield) in self._pre_evt_sel_static_data_fields_dict.items():
            dfield.calculate(self, src_hypo_group_manager, fitparams)

        self._trial_data_state_id += 1

    def calculate_static_data_fields(self, src_hypo_group_manager):
        """Calculates the data values of the data fields that do not depend on
        any source or fit parameters.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        """
        if(len(self._static_data_fields_dict) == 0):
            return

        fitparams = dict()
        for (name, dfield) in self._static_data_fields_dict.items():
            dfield.calculate(self, src_hypo_group_manager, fitparams)

        self._trial_data_state_id += 1

    def calculate_fitparam_data_fields(self, src_hypo_group_manager, fitparams):
        """Calculates the data values of the data fields that depend on fit
        parameter values.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager, which defines the groups of
            source hypotheses.
        fitparams : dict
            The dictionary holding the fit parameter names and values.
        """
        if(len(self._fitparam_data_fields_dict) == 0):
            return

        for (name, dfield) in self._fitparam_data_fields_dict.items():
            dfield.calculate(self, src_hypo_group_manager, fitparams)

        self._trial_data_state_id += 1

    def get_data(self, name):
        """Gets the data for the given data field name. The data is stored
        either in the raw events record ndarray or in one of the additional
        defined data fields. Data from the raw events record ndarray is
        prefered.

        Parameters
        ----------
        name : str
            The name of the data field for which to retrieve the data.

        Returns
        -------
        data : numpy ndarray
            The data of the requested data field.

        Raises
        ------
        KeyError
            If the given data field is not defined.
        """
        if((self._events is not None) and
           (name in self._events.field_name_list)):
            return self._events[name]

        if(name in self._source_data_fields_dict):
            data = self._source_data_fields_dict[name].values

            # Broadcast the value of an one-element 1D ndarray to the length
            # of the number of events. Note: Make sure that we don't broadcast
            # recarrays.
            if(self._events is not None):
                if((len(data) == 1) and (data.ndim == 1) and
                   (data.dtype.fields is None)):
                    data = np.repeat(data, len(self._events))
        else:
            raise KeyError(
                f'The data field "{name}" is not defined!')

        return data

    def get_dtype(self, name):
        """Gets the data type of the given data field.

        Parameters
        ----------
        name : str
            The name of the data field whose data type should get retrieved.

        Returns
        -------
        dt : numpy dtype
            The numpy dtype object of the given data field.

        Raises
        ------
        KeyError
            If the given data field is not defined.
        """
        dt = self.get_data(name).dtype

        return dt
